import logging
from pydantic import (
    BaseModel,
    BaseSettings,
    SecretStr,
    Field,
    PositiveInt,
    conint,
    constr,
    schema,
    validator,
    create_model,
)
from pydantic.dataclasses import dataclass
from pydantic.env_settings import SettingsSourceCallable
from pydantic.utils import ROOT_KEY
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)


logger = logging.getLogger(__name__)


class DynamicBaseModel(BaseModel):
    def __init__(__pydantic_self__, **data: Any) -> None:
        if __pydantic_self__.__custom_root_type__ and data.keys() != {ROOT_KEY}:
            data = {ROOT_KEY: data}
        super().__init__(**data)


class LabelStudioSecrets(BaseSettings):
    api_key: Optional[str] = SecretStr
    token: Optional[str] = SecretStr
    password: Optional[str] = SecretStr

    class Config:
        env_prefix = "LABELSTUDIO_"
        env_nested_delimiter = "_"
        case_sentive = False
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                env_settings,
                init_settings,
                file_secret_settings,
            )

class Secrets(BaseSettings):
    api_key: Optional[str] = SecretStr
    token: Optional[str] = SecretStr
    password: Optional[str] = SecretStr

    class Config:
        env_prefix = ""
        env_nested_delimiter = "_"
        case_sentive = False
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                env_settings,
                init_settings,
                file_secret_settings,
            )


class Environments(BaseSettings):
    EKORPKIT_CONFIG_DIR: Optional[str]
    EKORPKIT_DATA_DIR: Optional[str]
    EKORPKIT_PROJECT: Optional[str]
    EKORPKIT_WORKSPACE_ROOT: Optional[str]
    EKORPKIT_LOG_LEVEL: Optional[str]
    FRED_API_KEY: Optional[str] = SecretStr
    NASDAQ_API_KEY: Optional[str] = SecretStr
    WANDB_API_KEY: Optional[str] = SecretStr
    NUM_WORKERS: Optional[int]
    KMP_DUPLICATE_LIB_OK: Optional[str]
    CUDA_DEVICE_ORDER: Optional[str]
    CUDA_VISIBLE_DEVICES: Optional[str]

    class Config:
        env_prefix = ""
        case_sentive = False
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            return env_settings, file_secret_settings



from utils.utils import InputPadder
import numpy as np
import PIL, cv2
from PIL import Image
import torch


TAG_CHAR = np.array([202021.25], np.float32)


def writeFlow(filename, uv, v=None):
    """
    https://github.com/NVIDIA/flownet2-pytorch/blob/master/utils/flow_utils.py
    Copyright 2017 NVIDIA CORPORATION

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Write optical flow to file.

    If v is None, uv is assumed to contain both u and v channels,
    stacked in depth.
    Original code by Deqing Sun, adapted from Daniel Scharstein.
    """
    nBands = 2

    if v is None:
        assert uv.ndim == 3
        assert uv.shape[2] == 2
        u = uv[:, :, 0]
        v = uv[:, :, 1]
    else:
        u = uv

    assert u.shape == v.shape
    height, width = u.shape
    f = open(filename, "wb")
    # write the header
    f.write(TAG_CHAR)
    np.array(width).astype(np.int32).tofile(f)
    np.array(height).astype(np.int32).tofile(f)
    # arrange into matrix form
    tmp = np.zeros((height, width * nBands))
    tmp[:, np.arange(width) * 2] = u
    tmp[:, np.arange(width) * 2 + 1] = v
    tmp.astype(np.float32).tofile(f)
    f.close()


def load_img(img, size):
    img = Image.open(img).convert("RGB").resize(size)
    return torch.from_numpy(np.array(img)).permute(2, 0, 1).float()[None, ...].cuda()


def get_flow(frame1, frame2, model, iters=20):
    padder = InputPadder(frame1.shape)
    frame1, frame2 = padder.pad(frame1, frame2)
    _, flow12 = model(frame1, frame2, iters=iters, test_mode=True)
    flow12 = flow12[0].permute(1, 2, 0).detach().cpu().numpy()

    return flow12


def warp_flow(img, flow):
    h, w = flow.shape[:2]
    flow = flow.copy()
    flow[:, :, 0] += np.arange(w)
    flow[:, :, 1] += np.arange(h)[:, np.newaxis]
    res = cv2.remap(img, flow, None, cv2.INTER_LINEAR)
    return res


def makeEven(_x):
    return _x if (_x % 2 == 0) else _x + 1


def fit(img, maxsize=512):
    maxdim = max(*img.size)
    if maxdim > maxsize:
        # if True:
        ratio = maxsize / maxdim
        x, y = img.size
        size = (makeEven(int(x * ratio)), makeEven(int(y * ratio)))
        img = img.resize(size)
    return img


def warp(frame1, frame2, flo_path, blend=0.5, weights_path=None):
    flow21 = np.load(flo_path)
    frame1pil = np.array(
        frame1.convert("RGB").resize((flow21.shape[1], flow21.shape[0]))
    )
    frame1_warped21 = warp_flow(frame1pil, flow21)
    # frame2pil = frame1pil
    frame2pil = np.array(
        frame2.convert("RGB").resize((flow21.shape[1], flow21.shape[0]))
    )

    if weights_path:
        # TBD
        pass
    else:
        blended_w = frame2pil * (1 - blend) + frame1_warped21 * (blend)

    return PIL.Image.fromarray(blended_w.astype("uint8"))



import logging
import logging
import os
import pandas as pd
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class HFDS:
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self.autoload = self.args.get("autoload", True)
        self.name = self.args.get("name", None)
        self.subsets = self.args.get("subsets", None)
        self.splits = self.args.get("splits", None)
        self.verbose = self.args.get("verbose", True)
        self.download_mode = self.args.get("download_mode")
        self.ignore_verifications = self.args.get("ignore_verifications", True)

        self.output_dir = self.args.output_dir
        os.makedirs(self.output_dir, exist_ok=True)
        self.output_file = os.path.join(self.output_dir, self.args.output_file)
        self.force_download = self.args.force_download

        if self.autoload:
            self.build()

    def build(self):
        if not os.path.exists(self.output_file) or self.force_download:
            self.build_hfds()
        else:
            log.info(f"{self.output_file} already exists. skipping..")

    def build_hfds(self):
        from datasets import load_dataset

        dataset_name = self.name
        subsets = self.subsets
        splits = self.splits
        if isinstance(subsets, str):
            subsets = [subsets]
        elif not isinstance(subsets, list):
            subsets = [None]
        if isinstance(splits, str):
            splits = [splits]
        elif not isinstance(splits, list):
            subsets = [None]

        dfs = []
        for subset in subsets:
            for split in splits:
                ds = load_dataset(
                    dataset_name,
                    subset,
                    split=split,
                    download_mode=self.download_mode,
                    ignore_verifications=self.ignore_verifications,
                )
                print(ds)
                df = ds.to_pandas()
                df["subset"] = subset
                df[eKonf.Keys.SPLIT] = split
                dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)
        eKonf.save_data(df, self.output_file)
        if self.verbose:
            print(df.tail())
        log.info(f"Saved {len(df.index)} documents to {self.output_file}")



import os
import logging
import pandas as pd
from abc import ABCMeta, abstractmethod
from ekorpkit import eKonf
from .base import BaseSet


DESCRIPTION = "ekorpkit datasets"
LICENSE = "Copyright of the dataset is owned by the authors."


log = logging.getLogger(__name__)


class Corpus(BaseSet):
    def __init__(self, **args):
        super().__init__(**args)
        if isinstance(self.name, list):
            self.name = self.name[0]
        use_name_as_subdir = args.get("use_name_as_subdir", True)
        if use_name_as_subdir:
            self.data_dir = os.path.join(self.data_dir, self.name)
        self.metadata_dir = self.args.get("metadata_dir", None)
        if self.metadata_dir is None:
            self.metadata_dir = self.data_dir
        else:
            if use_name_as_subdir:
                self.metadata_dir = os.path.join(self.metadata_dir, self.name)

        self.load_info()
        self.load_column_info()

        self.meta_files = self.args.get("meta_files", None)
        self._metadata = None
        self._metadata_merged = False

        if self.auto.load:
            self.load()
            self.load_metadata()
            self.load_timestamp()
            if self.auto.merge:
                self.merge_metadata()

    @property
    def TEXT(self):
        return self.COLUMN.TEXT

    @property
    def METADATA(self):
        return self.COLUMN.METADATA

    @property
    def metadata(self):
        return self._metadata

    def load_timestamp(self):
        for split, data in self.splits.items():
            if data is None:
                continue
            data, self._metadata = self.COLUMN.to_timestamp(data, self.metadata)
            self._splits[split] = data
            if self.verbose:
                print(data.head())

    def load_metadata(self):
        if not self.meta_files or len(self.meta_files) == 0:
            log.info("No metadata files found")
            return

        dfs = []
        for split, data_file in self.meta_files.items():
            df = eKonf.load_data(data_file, self.metadata_dir, concatenate=True)
            df = self.COLUMN.append_split_to_meta(df, split)
            dfs.append(df)
        self._metadata = pd.concat(dfs)
        if self._collapse_ids:
            self._metadata = self.COLUMN.combine_ids(self._metadata)
        if self.verbose:
            log.info(f"Metadata loaded {len(self._metadata)} rows")
            print(self._metadata.head(3))
            print(self._metadata.tail(3))

    def merge_metadata(self):
        if self._metadata is None or self._metadata_merged:
            return
        for split, data in self.splits.items():
            if data is None:
                continue
            data = self.COLUMN.merge_metadata(data, self._metadata)
            data = self.COLUMN.init_info(data)
            self._splits[split] = data
        self._metadata_merged = True
        log.info(f"Metadata merged to data")
        if self.verbose:
            print(self._data.head(3))
            print(self._data.tail(3))



import os
import subprocess
import orjson as json
import pandas as pd
from multiprocessing import Pool
from ekorpkit import eKonf
from glob import glob
from ekorpkit.io.fetch.web import web_download, gdrive_download_un7z


class Wiki:
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self.name = self.args.name
        self.autoload = self.args.get("autoload", True)
        self.url = self.args.dump.url
        self.output_dir = self.args.output_dir
        self.dump_dir = self.args.dump.dump_dir
        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(self.dump_dir, exist_ok=True)
        if self.args.output_file:
            self.output_file = os.path.join(self.output_dir, self.args.output_file)
        self.dump_file = os.path.join(self.dump_dir, self.args.dump.dump_file)
        self.force_download = self.args.force_download

        if self.autoload:
            self.fetch()

    def fetch(self):
        if not os.listdir(self.output_dir) or self.force_download:
            self.download_dump()
            if self.args.extract._target_:
                getattr(self, self.args.extract._target_)()

    def extract_namuwiki(self):
        extracted_dir = self.dump_file[:-3]
        if os.path.exists(extracted_dir):
            json_file = glob(extracted_dir + "/*.json")[0]
        else:
            raise Exception("Extracted json file doesn't exist")

        with open(json_file, "r", encoding="utf-8") as input_file:
            namu_wiki = json.load(input_file)

        with Pool(processes=self.args.num_workers) as pool:
            documents = pool.map(work, namu_wiki)
        print(f"Extracted {self.name} from dump file {self.dump_file}")

        df = pd.DataFrame(documents)
        print(df.tail())
        df.to_csv(self.output_file, header=True, index=False, encoding="utf-8")
        print(f"Saved {self.name} to {self.output_file}")

    def extract_wiki(self):
        command = f"python -m wikiextractor.WikiExtractor {self.dump_file} --json --output {self.output_dir}"
        if self.args.num_workers > 1:
            command += f" --processes {self.args.num_workers}"
        if self.args.compress:
            command += " --compress"
        subprocess.run(command, shell=True)
        print(f"Extracted {self.name} from dump file {self.dump_file}")

    def download_dump(self):
        globals()[self.args.dump._target_](
            self.url, self.dump_file, self.name, self.force_download
        )


def work(document):
    from namuwiki.extractor import extract_text

    return {"title": document["title"], "text": extract_text(document["text"])}



import logging
import os
import sys


logger = logging.getLogger(__name__)


def _is_notebook():
    try:
        get_ipython
    except NameError:
        return False
    shell_type = get_ipython().__class__.__name__
    logger.info(f"shell type: {shell_type}")
    if shell_type == "ZMQInteractiveShell":
        return True  # Jupyter notebook or qtconsole
    elif shell_type == "Shell":
        return True  # Google colab
    elif shell_type == "TerminalInteractiveShell":
        return False  # Terminal running IPython
    else:
        return False  # Other type


def _is_colab():
    is_colab = "google.colab" in sys.modules
    if is_colab:
        logger.info("Google Colab detected.")
    else:
        logger.info("Google Colab not detected.")
    return is_colab


def _get_display():
    try:
        from ipywidgets import Output
    except ImportError:
        logger.info("ipywidgets not installed.")
        return None

    if _is_notebook():
        return Output()
    else:
        return None


def _clear_output(wait=False):
    from IPython import display

    if _is_notebook():
        display.clear_output(wait=True)


def _display(
    *objs,
    include=None,
    exclude=None,
    metadata=None,
    transient=None,
    display_id=None,
    raw=False,
    clear=False,
    **kwargs,
):
    from IPython import display

    if _is_notebook() and objs is not None:
        return display.display(
            *objs,
            include=include,
            exclude=exclude,
            metadata=metadata,
            transient=transient,
            display_id=display_id,
            raw=raw,
            clear=clear,
            **kwargs,
        )


def _display_image(
    data=None,
    url=None,
    filename=None,
    format=None,
    embed=None,
    width=None,
    height=None,
    retina=False,
    unconfined=False,
    metadata=None,
    **kwargs,
):
    from IPython import display

    if _is_notebook():
        img = display.Image(
            data=data,
            url=url,
            filename=filename,
            format=format,
            embed=embed,
            width=width,
            height=height,
            retina=retina,
            unconfined=unconfined,
            metadata=metadata,
            **kwargs,
        )
        return display.display(img)


def _hide_code_in_slideshow():
    from IPython import display
    import binascii

    uid = binascii.hexlify(os.urandom(8)).decode()
    html = """<div id="%s"></div>
    <script type="text/javascript">
        $(function(){
            var p = $("#%s");
            if (p.length==0) return;
            while (!p.hasClass("cell")) {
                p=p.parent();
                if (p.prop("tagName") =="body") return;
            }
            var cell = p;
            cell.find(".input").addClass("hide-in-slideshow")
        });
    </script>""" % (
        uid,
        uid,
    )
    display.display_html(html, raw=True)


def colored_str(s, color="black"):
    # return "<text style=color:{}>{}</text>".format(color, s)
    return "<text style=color:{}>{}</text>".format(color, s.replace("\n", "<br>"))


def _cprint(str_tuples):
    from IPython.display import HTML as html_print
    from IPython.display import display

    display(html_print(" ".join([colored_str(ti, color=ci) for ti, ci in str_tuples])))


def _create_dropdown(
    options,
    value,
    description,
    disabled=False,
    style={"description_width": "initial"},
    layout=None,
    **kwargs,
):
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    dropdown = widgets.Dropdown(
        options=options,
        value=value,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )
    return dropdown


def _create_textarea(
    value,
    description,
    placeholder="",
    disabled=False,
    style={"description_width": "initial"},
    layout=None,
    **kwargs,
):
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    textarea = widgets.Textarea(
        value=value,
        placeholder=placeholder,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )
    return textarea


def _create_button(description, button_style="", icon="check", layout=None, **kwargs):
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    button = widgets.Button(
        description=description,
        button_style=button_style,
        icon=icon,
        layout=layout,
        **kwargs,
    )
    return button


def _create_radiobutton(
    options,
    description,
    value=None,
    disabled=False,
    style={"description_width": "initial"},
    layout=None,
    **kwargs,
):
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    radiobutton = widgets.RadioButtons(
        options=options,
        value=value,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )
    return radiobutton


def _create_image(
    filename=None,
    format=None,
    width=None,
    height=None,
    **kwargs,
):
    import ipywidgets as widgets

    # from urllib.request import urlopen

    if filename is None:
        url = "https://github.com/entelecheia/ekorpkit-book/raw/main/assets/figs/placeholder.png"
        # img = urlopen(url).read()
        img = _read(url)
        format = "png"
    else:
        img = _read(filename)
        format = format or filename.split(".")[-1]
    image = widgets.Image(
        value=img,
        format=format,
        width=width,
        height=height,
        **kwargs,
    )
    return image


def _create_floatslider(
    min=0.0,
    max=1.0,
    step=0.1,
    value=None,
    description="",
    disabled=False,
    continuous_update=False,
    orientation="horizontal",
    readout=True,
    readout_format=".1f",
    style={"description_width": "initial"},
    layout=None,
    **kwargs,
):
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    slider = widgets.FloatSlider(
        min=min,
        max=max,
        step=step,
        value=value,
        description=description,
        disabled=disabled,
        continuous_update=continuous_update,
        orientation=orientation,
        readout=readout,
        readout_format=readout_format,
        style=style,
        layout=layout,
        **kwargs,
    )
    return slider


def _read(uri, mode="rb", encoding=None, **kwargs):
    if uri.startswith("http"):
        import requests

        return requests.get(uri, **kwargs).content
    # elif uri.startswith("s3://"):
    #     import boto3

    #     s3 = boto3.resource("s3")
    #     bucket, key = uri.replace("s3://", "").split("/", 1)
    #     obj = s3.Object(bucket, key)
    #     return obj.get()["Body"].read()
    else:
        with open(uri, mode=mode, encoding=encoding) as f:
            return f.read()



import click
import cssutils
import json
import logging
import numpy as np
import os
import pandas as pd
import re
import sys
from bs4 import BeautifulSoup
from html.parser import HTMLParser
from pathos.multiprocessing import ProcessPool
from tqdm.auto import tqdm
from typing import List
from ekorpkit import eKonf

# Change the default recursion limit of 1000 to 30000
sys.setrecursionlimit(30000)

# Supress cssutils stupid warnings
cssutils.log.setLevel(logging.CRITICAL)

cli = click.Group()

regex_flags = re.IGNORECASE | re.DOTALL | re.MULTILINE

# Instantiate a logger object
log = logging.getLogger(__name__)


class HtmlStripper(HTMLParser):
    """
    Strips HTML tags
    """

    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return "".join(self.fed)

    def strip_tags(self, html):
        self.feed(html)
        return self.get_data()


class ExtractItems:
    def __init__(
        self,
        remove_tables: bool,
        items_to_extract: List,
        raw_files_dir: str,
        extracted_files_dir: str,
    ):

        self.remove_tables = remove_tables
        self.items_list = [
            "1",
            "1A",
            "1B",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "7A",
            "8",
            "9",
            "9A",
            "9B",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
        ]
        self.items_to_extract = (
            items_to_extract if items_to_extract else self.items_list
        )
        self.raw_files_folder = raw_files_dir
        self.extracted_files_folder = extracted_files_dir

    @staticmethod
    def strip_html(html_content):
        """
        Strips the html content to get clean text
        :param html_content: The HTML content
        :return: The clean HTML content
        """

        # TODO: Check if flags are required in the following regex
        html_content = re.sub(r"(<\s*/\s*(div|tr|p|li|)\s*>)", r"\1\n\n", html_content)
        html_content = re.sub(r"(<br\s*>|<br\s*/>)", r"\1\n\n", html_content)
        html_content = re.sub(r"(<\s*/\s*(th|td)\s*>)", r" \1 ", html_content)
        html_content = HtmlStripper().strip_tags(html_content)

        return html_content

    @staticmethod
    def remove_multiple_lines(text):
        """
        Replaces consecutive new lines with a single new line
        and consecutive whitespace characters with a single whitespace
        :param text: String containing the financial text
        :return: String without multiple newlines
        """

        text = re.sub(r"(( )*\n( )*){2,}", "#NEWLINE", text)
        text = re.sub(r"\n", " ", text)
        text = re.sub(r"(#NEWLINE)+", "\n", text).strip()
        text = re.sub(r"[ ]{2,}", " ", text)

        return text

    @staticmethod
    def clean_text(text):
        """
        Clean the text of various unnecessary blocks of text
        Substitute various special characters

        :param text: Raw text string
        :return: String containing normalized, clean text
        """

        text = re.sub(r"[\xa0]", " ", text)
        text = re.sub(r"[\u200b]", " ", text)

        text = re.sub(r"[\x91]", "‘", text)
        text = re.sub(r"[\x92]", "’", text)
        text = re.sub(r"[\x93]", "“", text)
        text = re.sub(r"[\x94]", "”", text)
        text = re.sub(r"[\x95]", "•", text)
        text = re.sub(r"[\x96]", "-", text)
        text = re.sub(r"[\x97]", "-", text)
        text = re.sub(r"[\x98]", "˜", text)
        text = re.sub(r"[\x99]", "™", text)

        text = re.sub(r"[\u2010\u2011\u2012\u2013\u2014\u2015]", "-", text)

        def remove_whitespace(match):
            ws = r"[^\S\r\n]"
            return f'{match[1]}{re.sub(ws, r"", match[2])}{match[3]}{match[4]}'

        # Fix broken section headers
        text = re.sub(
            r"(\n[^\S\r\n]*)(P[^\S\r\n]*A[^\S\r\n]*R[^\S\r\n]*T)([^\S\r\n]+)((\d{1,2}|[IV]{1,2})[AB]?)",
            remove_whitespace,
            text,
            flags=re.IGNORECASE,
        )
        text = re.sub(
            r"(\n[^\S\r\n]*)(I[^\S\r\n]*T[^\S\r\n]*E[^\S\r\n]*M)([^\S\r\n]+)(\d{1,2}[AB]?)",
            remove_whitespace,
            text,
            flags=re.IGNORECASE,
        )

        text = re.sub(
            r"(ITEM|PART)(\s+\d{1,2}[AB]?)([\-•])",
            r"\1\2 \3 ",
            text,
            flags=re.IGNORECASE,
        )

        # Remove unnecessary headers
        text = re.sub(
            r"\n[^\S\r\n]*"
            r"(TABLE\s+OF\s+CONTENTS|INDEX\s+TO\s+FINANCIAL\s+STATEMENTS|BACK\s+TO\s+CONTENTS|QUICKLINKS)"
            r"[^\S\r\n]*\n",
            "\n",
            text,
            flags=regex_flags,
        )

        # Remove page numbers and headers
        text = re.sub(
            r"\n[^\S\r\n]*[-‒–—]*\d+[-‒–—]*[^\S\r\n]*\n", "\n", text, flags=regex_flags
        )
        text = re.sub(r"\n[^\S\r\n]*\d+[^\S\r\n]*\n", "\n", text, flags=regex_flags)

        text = re.sub(r"[\n\s]F[-‒–—]*\d+", "", text, flags=regex_flags)
        text = re.sub(
            r"\n[^\S\r\n]*Page\s[\d*]+[^\S\r\n]*\n", "", text, flags=regex_flags
        )

        return text

    @staticmethod
    def calculate_table_character_percentages(table_text):
        """
        Calculate character type percentages contained in the table text

        :param table_text: The table text
        :return non_blank_digits_percentage: Percentage of digit characters
        :return spaces_percentage: Percentage of space characters
        """
        digits = sum(c.isdigit() for c in table_text)
        # letters   = sum(c.isalpha() for c in table_text)
        spaces = sum(c.isspace() for c in table_text)

        if len(table_text) - spaces:
            non_blank_digits_percentage = digits / (len(table_text) - spaces)
        else:
            non_blank_digits_percentage = 0

        if len(table_text):
            spaces_percentage = spaces / len(table_text)
        else:
            spaces_percentage = 0

        return non_blank_digits_percentage, spaces_percentage

    @staticmethod
    def remove_html_tables(doc_10k, is_html):
        """
        Remove HTML tables that contain numerical data
        Note that there are many corner-cases in the tables that have text data instead of numerical

        :param doc_10k: The 10-K html
        :param is_html: Whether the document contains html code or just plain text
        :return: doc_10k: The 10-K html without numerical tables
        """

        if is_html:
            tables = doc_10k.find_all("table")

            # Detect tables that have numerical data
            for tbl in tables:
                trs = (
                    tbl.find_all("tr", attrs={"style": True})
                    + tbl.find_all("td", attrs={"style": True})
                    + tbl.find_all("th", attrs={"style": True})
                )

                background_found = False

                for tr in trs:
                    # Parse given cssText which is assumed to be the content of a HTML style attribute
                    style = cssutils.parseStyle(tr["style"])

                    if (
                        style["background"]
                        and style["background"].lower()
                        not in ["none", "transparent", "#ffffff", "#fff", "white"]
                    ) or (
                        style["background-color"]
                        and style["background-color"].lower()
                        not in ["none", "transparent", "#ffffff", "#fff", "white"]
                    ):
                        background_found = True
                        break

                trs = (
                    tbl.find_all("tr", attrs={"bgcolor": True})
                    + tbl.find_all("td", attrs={"bgcolor": True})
                    + tbl.find_all("th", attrs={"bgcolor": True})
                )

                bgcolor_found = False
                for tr in trs:
                    if tr["bgcolor"].lower() not in [
                        "none",
                        "transparent",
                        "#ffffff",
                        "#fff",
                        "white",
                    ]:
                        bgcolor_found = True
                        break

                if bgcolor_found or background_found:
                    tbl.decompose()

        else:
            doc_10k = re.sub(r"<TABLE>.*?</TABLE>", "", str(doc_10k), flags=regex_flags)

        return doc_10k

    def parse_item(self, text, item_index, next_item_list, positions):
        """
        Parses Item N for a 10-K text

        :param text: The 10-K text
        :param item_index: Number of the requested Item/Section of the 10-K text
        :param next_item_list: List of possible next 10-K item sections
        :param positions: List of the end positions of previous item sections
        :return: item_section: The item/section as a text string
        """

        if item_index == "9A":
            item_index = item_index.replace("A", r"[^\S\r\n]*A(?:\(T\))?")
        elif "A" in item_index:
            item_index = item_index.replace("A", r"[^\S\r\n]*A")
        elif "B" in item_index:
            item_index = item_index.replace("B", r"[^\S\r\n]*B")

        # Depending on the item_index, search for subsequent sections.

        # There might be many 'candidate' text sections between 2 Items.
        # For example, the Table of Contents (ToC) still counts as a match when searching text between 'Item 3' and 'Item 4'
        # But we do NOT want that specific text section; We want the detailed section which is *after* the ToC

        possible_sections_list = []
        for next_item_index in next_item_list:
            if possible_sections_list:
                break
            if next_item_index == "9A":
                next_item_index = next_item_index.replace("A", r"[^\S\r\n]*A(?:\(T\))?")
            elif "A" in next_item_index:
                next_item_index = next_item_index.replace("A", r"[^\S\r\n]*A")
            elif "B" in next_item_index:
                next_item_index = next_item_index.replace("B", r"[^\S\r\n]*B")

            for match in list(
                re.finditer(
                    rf"\n[^\S\r\n]*ITEM\s+{item_index}[.*~\-:\s]",
                    text,
                    flags=regex_flags,
                )
            ):
                offset = match.start()

                possible = list(
                    re.finditer(
                        rf"\n[^\S\r\n]*ITEM\s+{item_index}[.*~\-:\s].+?([^\S\r\n]*ITEM\s+{str(next_item_index)}[.*~\-:\s])",
                        text[offset:],
                        flags=regex_flags,
                    )
                )
                if possible:
                    possible_sections_list += [(offset, possible)]

        # Extract the wanted section from the text
        item_section, positions = ExtractItems.get_item_section(
            possible_sections_list, text, positions
        )

        # If item is the last one (usual case when dealing with EDGAR's old .txt files), get all the text from its beginning until EOF.
        if positions:
            if item_index in self.items_list and item_section == "":
                item_section = ExtractItems.get_last_item_section(
                    item_index, text, positions
                )
            elif (
                item_index == "15"
            ):  # Item 15 is the last one, get all the text from its beginning until EOF
                item_section = ExtractItems.get_last_item_section(
                    item_index, text, positions
                )

        return item_section.strip(), positions

    @staticmethod
    def get_item_section(possible_sections_list, text, positions):
        """
        Throughout a list of all the possible item sections, it returns the biggest one, which (probably) is the correct one.

        :param possible_sections_list: List containing all the possible sections betweewn Item X and Item Y
        :param text: The whole text
        :param positions: List of the end positions of previous item sections
        :return: The correct section
        """

        item_section = ""
        max_match_length = 0
        max_match = None
        max_match_offset = None

        # Find the match with the largest section
        for (offset, matches) in possible_sections_list:
            for match in matches:
                match_length = match.end() - match.start()
                if positions:
                    if (
                        match_length > max_match_length
                        and offset + match.start() >= positions[-1]
                    ):
                        max_match = match
                        max_match_offset = offset
                        max_match_length = match_length
                elif match_length > max_match_length:
                    max_match = match
                    max_match_offset = offset
                    max_match_length = match_length

        # Return the text section inside that match
        if max_match:
            if positions:
                if max_match_offset + max_match.start() >= positions[-1]:
                    item_section = text[
                        max_match_offset
                        + max_match.start() : max_match_offset
                        + max_match.regs[1][0]
                    ]
            else:
                item_section = text[
                    max_match_offset
                    + max_match.start() : max_match_offset
                    + max_match.regs[1][0]
                ]
            positions.append(max_match_offset + max_match.end() - len(max_match[1]) - 1)

        return item_section, positions

    @staticmethod
    def get_last_item_section(item_index, text, positions):
        """
        Returns the text section starting through a given item. This is useful in cases where Item 15 is the last item
        and there is no Item 16 to indicate its ending. Also, it is useful in cases like EDGAR's old .txt files
        (mostly before 2005), where there there is no Item 15; thus, ITEM 14 is the last one there.

        :param item_index: The index of the item/section in the 10-K ('14' or '15')
        :param text: The whole 10-K text
        :param positions: List of the end positions of previous item sections
        :return: All the remaining text until the end, starting from the specified item_index
        """

        item_list = list(
            re.finditer(
                rf"\n[^\S\r\n]*ITEM\s+{item_index}[.\-:\s].+?", text, flags=regex_flags
            )
        )

        item_section = ""
        for item in item_list:
            if item.start() >= positions[-1]:
                item_section = text[item.start() :].strip()
                break

        return item_section

    def extract_items(self, filing_metadata):
        """
        Extracts all items/sections for a 10-K file and writes it to a CIK_10K_YEAR.json file (eg. 1384400_10K_2017.json)

        :param filing_metadata: a pandas series containing all filings metadata
        """

        absolute_10k_filename = os.path.join(
            self.raw_files_folder, filing_metadata["filename"]
        )

        with open(absolute_10k_filename, "r", errors="backslashreplace") as file:
            content = file.read()

        # Remove all embedded pdfs that might be seen in few old 10-K txt annual reports
        content = re.sub(r"<PDF>.*?</PDF>", "", content, flags=regex_flags)

        documents = re.findall("<DOCUMENT>.*?</DOCUMENT>", content, flags=regex_flags)

        doc_10k = None
        found_10k, is_html = False, False
        for doc in documents:
            doc_type = re.search(r"\n[^\S\r\n]*<TYPE>(.*?)\n", doc, flags=regex_flags)
            doc_type = doc_type.group(1) if doc_type else None
            if doc_type.startswith("10"):
                doc_10k = BeautifulSoup(doc, "lxml")
                is_html = (True if doc_10k.find("td") else False) and (
                    True if doc_10k.find("tr") else False
                )
                if not is_html:
                    doc_10k = doc
                found_10k = True
                break

        if not found_10k:
            if documents:
                log.info(
                    f'Could not find document type 10K for {filing_metadata["filename"]}'
                )
            doc_10k = BeautifulSoup(content, "lxml")
            is_html = (True if doc_10k.find("td") else False) and (
                True if doc_10k.find("tr") else False
            )
            if not is_html:
                doc_10k = content

        # if not is_html and not documents:
        if filing_metadata["filename"].endswith("txt") and not documents:
            log.info(f'No <DOCUMENT> tag for {filing_metadata["filename"]}')

        # For non html clean all table items
        if self.remove_tables:
            doc_10k = self.remove_html_tables(doc_10k, is_html=is_html)

        json_content = {
            "cik": filing_metadata["CIK"],
            "company": filing_metadata["Company"],
            "filing_type": filing_metadata["Type"],
            "filing_date": filing_metadata["Date"],
            "period_of_report": filing_metadata["Period of Report"],
            "sic": filing_metadata["SIC"],
            "state_of_inc": filing_metadata["State of Inc"],
            "state_location": filing_metadata["State location"],
            "fiscal_year_end": filing_metadata["Fiscal Year End"],
            "filing_html_index": filing_metadata["html_index"],
            "htm_filing_link": filing_metadata["htm_file_link"],
            "complete_text_filing_link": filing_metadata["complete_text_file_link"],
            "filename": filing_metadata["filename"],
        }
        for item_index in self.items_to_extract:
            json_content[f"item_{item_index}"] = ""

        text = ExtractItems.strip_html(str(doc_10k))
        text = ExtractItems.clean_text(text)

        positions = []
        all_items_null = True
        for i, item_index in enumerate(self.items_list):
            next_item_list = self.items_list[i + 1 :]
            item_section, positions = self.parse_item(
                text, item_index, next_item_list, positions
            )
            item_section = ExtractItems.remove_multiple_lines(item_section)

            if item_index in self.items_to_extract:
                if item_section != "":
                    all_items_null = False
                json_content[f"item_{item_index}"] = item_section

        if all_items_null:
            log.info(f"Could not extract any item for {absolute_10k_filename}")
            return None

        return json_content

    def process_filing(self, filing_metadata):
        json_filename = f'{filing_metadata["filename"].split(".")[0]}.json'
        absolute_json_filename = os.path.join(
            self.extracted_files_folder, json_filename
        )
        if os.path.exists(absolute_json_filename):
            return 0

        json_content = self.extract_items(filing_metadata)

        if json_content is not None:
            with open(absolute_json_filename, "w") as filepath:
                json.dump(json_content, filepath, indent=4)

        return 1


def extract_items(**args):
    """
    Gets the list of 10K files and extracts all textual items/sections by calling the extract_items() function.
    """

    config = eKonf.to_config(args)
    output_dir = config["output_dir"]

    filings_metadata_filepath = os.path.join(
        output_dir, config["filings_metadata_file"]
    )
    if os.path.exists(filings_metadata_filepath):
        filings_metadata_df = pd.read_csv(filings_metadata_filepath, dtype=str)
        filings_metadata_df = filings_metadata_df.replace({np.nan: None})

    else:
        log.info(f'No such file "{filings_metadata_filepath}"')
        return

    raw_filings_dir = os.path.join(output_dir, config["raw_filings_dir"])
    if not os.path.isdir(raw_filings_dir):
        log.info(f'No such directory: "{raw_filings_dir}')
        return

    extracted_filings_dir = os.path.join(output_dir, config["extracted_filings_dir"])

    if not os.path.isdir(extracted_filings_dir):
        os.mkdir(extracted_filings_dir)

    extraction = ExtractItems(
        remove_tables=config["remove_tables"],
        items_to_extract=config["items_to_extract"],
        raw_files_dir=raw_filings_dir,
        extracted_files_dir=extracted_filings_dir,
    )

    log.info(f"Starting extraction...\n")

    list_of_series = list(zip(*filings_metadata_df.iterrows()))[1]

    with ProcessPool(processes=1) as pool:
        processed = list(
            tqdm(
                pool.imap(extraction.process_filing, list_of_series),
                total=len(list_of_series),
                ncols=100,
            )
        )

    log.info(f"\nItem extraction is completed successfully.")
    log.info(f"{sum(processed)} files were processed.")
    log.info(f"Extracted filings are saved to: {extracted_filings_dir}")



import logging
import os
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from ..analysis import Scorer
from .data import DictDataLoader
from .utils import metrics_dict_to_dataframe, move_to_device
from ..types import Config
from ..utils import probs_to_preds

from .task import Operation, Task

OutputDict = Dict[str, Union[Any, Mapping[str, Any]]]


class ClassifierConfig(Config):
    """A classifier built from one or more tasks to support advanced workflows.

    Parameters
    ----------
    device
        The device (GPU) to move the model to (-1 is CPU), but device will also be
        moved to CPU if no GPU device is available
    dataparallel
        Whether or not to use PyTorch DataParallel wrappers to automatically utilize
        multiple GPUs if available
    """

    device: int = 0
    dataparallel: bool = True


class MultitaskClassifier(nn.Module):
    r"""A classifier built from one or more tasks to support advanced workflows.

    Parameters
    ----------
    tasks
        A list of ``Task``\s to build a model from
    name
        The name of the classifier

    Attributes
    ----------
    config
        The config dict containing the settings for this model
    name
        See above
    module_pool
        A dictionary of all modules used by any of the tasks (See Task docstring)
    task_names
        See Task docstring
    op_sequences
        See Task docstring
    loss_funcs
        See Task docstring
    output_funcs
        See Task docstring
    scorers
        See Task docstring
    """

    def __init__(
        self, tasks: List[Task], name: Optional[str] = None, **kwargs: Any
    ) -> None:
        super().__init__()
        self.config = ClassifierConfig(**kwargs)
        self.name = name or type(self).__name__

        # Initiate the model attributes
        self.module_pool = nn.ModuleDict()
        self.task_names: Set[str] = set()
        self.op_sequences: Dict[str, Sequence[Operation]] = dict()
        self.loss_funcs: Dict[str, Callable[..., torch.Tensor]] = dict()
        self.output_funcs: Dict[str, Callable[..., torch.Tensor]] = dict()
        self.scorers: Dict[str, Scorer] = dict()

        # Build network with given tasks
        self._build_network(tasks)

        # Report total task count and duplicates
        all_ops = [op.name for t in tasks for op in t.op_sequence]
        unique_ops = set(all_ops)
        all_mods = [mod_name for t in tasks for mod_name in t.module_pool]
        unique_mods = set(all_mods)
        logging.info(
            f"Created multi-task model {self.name} that contains "
            f"task(s) {self.task_names} from "
            f"{len(unique_ops)} operations ({len(all_ops) - len(unique_ops)} shared) and "
            f"{len(unique_mods)} modules ({len(all_mods) - len(unique_mods)} shared)."
        )

        # Move model to specified device
        self._move_to_device()

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"

    def _build_network(self, tasks: List[Task]) -> None:
        r"""Construct the network from a list of ``Task``\s by adding them one by one.

        Parameters
        ----------
        tasks
            A list of ``Task``s
        """
        for task in tasks:
            if not isinstance(task, Task):
                raise ValueError(f"Unrecognized task type {task}.")
            if task.name in self.task_names:
                raise ValueError(
                    f"Found duplicate task {task.name}, different task should use "
                    f"different task name."
                )
            self.add_task(task)

    def add_task(self, task: Task) -> None:
        """Add a single task to the network.

        Parameters
        ----------
        task
            A ``Task`` to add
        """
        # Combine module_pool from all tasks
        for key in task.module_pool.keys():
            if key in self.module_pool.keys():
                if self.config.dataparallel:
                    task.module_pool[key] = nn.DataParallel(self.module_pool[key])
                else:
                    task.module_pool[key] = self.module_pool[key]
            else:
                if self.config.dataparallel:
                    self.module_pool[key] = nn.DataParallel(task.module_pool[key])
                else:
                    self.module_pool[key] = task.module_pool[key]
        self.task_names.add(task.name)
        self.op_sequences[task.name] = task.op_sequence
        self.loss_funcs[task.name] = task.loss_func
        self.output_funcs[task.name] = task.output_func
        self.scorers[task.name] = task.scorer

        # Move model to specified device
        self._move_to_device()

    def forward(  # type: ignore
        self, X_dict: Dict[str, Any], task_names: Iterable[str]
    ) -> OutputDict:
        """Do a forward pass through the network for all specified tasks.

        Parameters
        ----------
        X_dict
            A dict of data fields
        task_names
            The names of the tasks to execute the forward pass for

        Returns
        -------
        OutputDict
            A dict mapping each operation name to its corresponding output

        Raises
        ------
        TypeError
            If an Operation input has an invalid type
        ValueError
            If a specified Operation failed to execute
        """
        X_dict_moved = move_to_device(X_dict, self.config.device)

        outputs: OutputDict = {"_input_": X_dict_moved}  # type: ignore

        # Call forward for each task, using cached result if available
        # Each op_sequence consists of one or more operations that are executed in order
        for task_name in task_names:
            op_sequence = self.op_sequences[task_name]

            for operation in op_sequence:
                if operation.name not in outputs:
                    try:
                        if operation.inputs:
                            # Feed the inputs the module requested in the reqested order
                            inputs = []
                            for op_input in operation.inputs:
                                if isinstance(op_input, tuple):
                                    # The output of the indicated operation has a dict
                                    # of fields; extract the designated field by name
                                    op_name, field_key = op_input
                                    inputs.append(outputs[op_name][field_key])
                                else:
                                    # The output of the indicated operation has only
                                    # one field; use that as the input to the current op
                                    op_name = op_input
                                    inputs.append(outputs[op_name])

                            output = self.module_pool[operation.module_name].forward(
                                *inputs
                            )
                        else:
                            # Feed the entire outputs dict for the module to pull from
                            output = self.module_pool[operation.module_name].forward(
                                outputs
                            )
                    except Exception as e:
                        raise ValueError(
                            f"Unsuccessful operation {operation}: {repr(e)}."
                        )
                    outputs[operation.name] = output

        return outputs

    def calculate_loss(
        self, X_dict: Dict[str, Any], Y_dict: Dict[str, torch.Tensor]
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, float]]:
        """Calculate the loss for each task and the number of data points contributing.

        Parameters
        ----------
        X_dict
            A dict of data fields
        Y_dict
            A dict from task names to label sets

        Returns
        -------
        Dict[str, torch.Tensor], Dict[str, float]
            A dict of losses by task name and seen examples by task name
        """

        loss_dict = dict()
        count_dict = dict()

        labels_to_tasks = self._get_labels_to_tasks(Y_dict.keys())
        outputs = self.forward(X_dict, task_names=labels_to_tasks.values())

        # Calculate loss for each task
        for label_name, task_name in labels_to_tasks.items():
            Y = Y_dict[label_name]

            # Select the active samples
            if len(Y.size()) == 1:
                active = Y.detach() != -1
            else:
                active = torch.any(Y.detach() != -1, dim=1)

            # Only calculate the loss when active example exists
            if active.any():
                # Note: Use label_name as key, but task_name to access model attributes
                count_dict[label_name] = active.sum().item()

                # Extract the output of the last operation for this task
                inputs = outputs[self.op_sequences[task_name][-1].name]

                # Filter out any inactive examples if inputs is a Tensor
                if not active.all() and isinstance(inputs, torch.Tensor):
                    inputs = inputs[active]
                    Y = Y[active]

                loss_dict[label_name] = self.loss_funcs[task_name](
                    inputs, move_to_device(Y, self.config.device)
                )

        return loss_dict, count_dict

    @torch.no_grad()
    def _calculate_probs(
        self, X_dict: Dict[str, Any], task_names: Iterable[str]
    ) -> Dict[str, Iterable[torch.Tensor]]:
        """Calculate the probabilities for each task.

        Parameters
        ----------
        X_dict
            A dict of data fields
        task_names
            A list of task names to calculate probabilities for

        Returns
        -------
        Dict[str, Iterable[torch.Tensor]]
            A dictionary mapping task name to probabilities
        """

        self.eval()

        prob_dict = dict()

        outputs = self.forward(X_dict, task_names)

        for task_name in task_names:
            # Extract the output of the last operation for this task
            inputs = outputs[self.op_sequences[task_name][-1].name]
            prob_dict[task_name] = self.output_funcs[task_name](inputs).cpu().numpy()

        return prob_dict

    @torch.no_grad()
    def predict(
        self,
        dataloader: DictDataLoader,
        return_preds: bool = False,
        remap_labels: Dict[str, Optional[str]] = {},
    ) -> Dict[str, Dict[str, torch.Tensor]]:
        """Calculate probabilities, (optionally) predictions, and pull out gold labels.

        Parameters
        ----------
        dataloader
            A DictDataLoader to make predictions for
        return_preds
            If True, include predictions in the return dict (not just probabilities)
        remap_labels
            A dict specifying which labels in the dataset's Y_dict (key)
            to remap to a new task (value)

        Returns
        -------
        Dict[str, Dict[str, torch.Tensor]]
            A dictionary mapping label type ('golds', 'probs', 'preds') to values
        """
        self.eval()

        gold_dict_list: Dict[str, List[torch.Tensor]] = defaultdict(list)
        prob_dict_list: Dict[str, List[torch.Tensor]] = defaultdict(list)

        labels_to_tasks = self._get_labels_to_tasks(
            label_names=dataloader.dataset.Y_dict.keys(),  # type: ignore
            remap_labels=remap_labels,
        )
        for batch_num, (X_batch_dict, Y_batch_dict) in enumerate(dataloader):
            prob_batch_dict = self._calculate_probs(
                X_batch_dict, labels_to_tasks.values()
            )
            for label_name in labels_to_tasks:
                task_name = labels_to_tasks[label_name]
                Y = Y_batch_dict[label_name]

                # Note: store results under label_name
                # but retrieve from pre-computed results using task_name
                prob_dict_list[label_name].extend(prob_batch_dict[task_name])
                gold_dict_list[label_name].extend(Y.cpu().numpy())

        gold_dict: Dict[str, torch.Tensor] = {}
        prob_dict: Dict[str, torch.Tensor] = {}

        for task_name in gold_dict_list:
            gold_dict[task_name] = torch.Tensor(np.array(gold_dict_list[task_name]))
            prob_dict[task_name] = torch.Tensor(np.array(prob_dict_list[task_name]))

        if return_preds:
            pred_dict: Dict[str, torch.Tensor] = defaultdict(torch.Tensor)
            for task_name, probs in prob_dict.items():
                pred_dict[task_name] = torch.Tensor(probs_to_preds(probs.numpy()))

        results = {"golds": gold_dict, "probs": prob_dict}

        if return_preds:
            results["preds"] = pred_dict

        return results

    @torch.no_grad()
    def score(
        self,
        dataloaders: List[DictDataLoader],
        remap_labels: Dict[str, Optional[str]] = {},
        as_dataframe: bool = False,
    ) -> Union[Dict[str, float], pd.DataFrame]:
        """Calculate scores for the provided DictDataLoaders.

        Parameters
        ----------
        dataloaders
            A list of DictDataLoaders to calculate scores for
        remap_labels
            A dict specifying which labels in the dataset's Y_dict (key)
            to remap to a new task (value)
        as_dataframe
            A boolean indicating whether to return results as pandas
            DataFrame (True) or dict (False)

        Returns
        -------
        Dict[str, float]
            A dictionary mapping metric names to corresponding scores
            Metric names will be of the form "task/dataset/split/metric"
        """

        self.eval()

        metric_score_dict = dict()

        for dataloader in dataloaders:
            # Construct label to task mapping for evaluation
            Y_dict = dataloader.dataset.Y_dict  # type: ignore
            labels_to_tasks = self._get_labels_to_tasks(
                label_names=Y_dict.keys(), remap_labels=remap_labels
            )

            # What labels in Y_dict are we ignoring?
            extra_labels = set(Y_dict.keys()).difference(set(labels_to_tasks.keys()))
            if extra_labels:
                logging.info(
                    f"Ignoring extra labels in dataloader ({dataloader.dataset.split}): {extra_labels}"  # type: ignore
                )

            # Obtain predictions
            results = self.predict(
                dataloader, return_preds=True, remap_labels=remap_labels
            )
            # Score and record metrics for each set of predictions
            for label_name, task_name in labels_to_tasks.items():
                metric_scores = self.scorers[task_name].score(
                    golds=results["golds"][label_name].numpy(),
                    preds=results["preds"][label_name].numpy(),
                    probs=results["probs"][label_name].numpy(),
                )

                for metric_name, metric_value in metric_scores.items():
                    # Type ignore statements are necessary because the DataLoader class
                    # that DictDataLoader inherits from is what actually sets
                    # the class of Dataset, and it doesn't know about name and split.
                    identifier = "/".join(
                        [
                            label_name,
                            dataloader.dataset.name,  # type: ignore
                            dataloader.dataset.split,  # type: ignore
                            metric_name,
                        ]
                    )
                    metric_score_dict[identifier] = metric_value

        if as_dataframe:
            return metrics_dict_to_dataframe(metric_score_dict)

        return metric_score_dict

    def _get_labels_to_tasks(
        self, label_names: Iterable[str], remap_labels: Dict[str, Optional[str]] = {}
    ) -> Dict[str, str]:
        """Map each label to its corresponding task outputs based on whether the task is available.

        If remap_labels specified, overrides specific label -> task mappings.
        If a label is mappied to `None`, that key is removed from the mapping.
        """
        labels_to_tasks = {}
        for label in label_names:
            # Override any existing label -> task mappings
            if label in remap_labels:
                task = remap_labels.get(label)
                # Note: task might be manually remapped to None to remove it from the labels_to_tasks
                if task is not None:
                    labels_to_tasks[label] = task

            # If available in task flows, label should map to task of same name
            elif label in self.op_sequences:
                labels_to_tasks[label] = label

        return labels_to_tasks

    def _move_to_device(self) -> None:  # pragma: no cover
        """Move the model to the device specified in the model config."""
        device = self.config.device
        if device >= 0:
            if torch.cuda.is_available():
                logging.info(f"Moving model to GPU (cuda:{device}).")
                self.to(torch.device(f"cuda:{device}"))
            else:
                logging.info("No cuda device available. Switch to cpu instead.")

    def save(self, model_path: str) -> None:
        """Save the model to the specified file path.

        Parameters
        ----------
        model_path
            The path where the model should be saved

        Raises
        ------
        BaseException
            If the torch.save() method fails
        """
        if not os.path.exists(os.path.dirname(model_path)):
            os.makedirs(os.path.dirname(model_path))

        try:
            torch.save(self.state_dict(), model_path)
        except BaseException:  # pragma: no cover
            logging.warning("Saving failed... continuing anyway.")

        logging.info(f"[{self.name}] Model saved in {model_path}")

    def load(self, model_path: str) -> None:
        """Load a saved model from the provided file path and moves it to a device.

        Parameters
        ----------
        model_path
            The path to a saved model
        """
        try:
            self.load_state_dict(
                torch.load(model_path, map_location=torch.device("cpu"))
            )
        except BaseException:  # pragma: no cover
            if not os.path.exists(model_path):
                logging.error("Loading failed... Model does not exist.")
            else:
                logging.error(f"Loading failed... Cannot load model from {model_path}")
            raise

        logging.info(f"[{self.name}] Model loaded from {model_path}")
        self._move_to_device()



# Korean Sentence Splitter
# Split Korean text into sentences using heuristic algorithm.
#
# Copyright (C) 2021 Hyun-Woong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from collections import defaultdict
from typing import Any


class Stats(object):
    DEFAULT: int = 0
    DA_EOJEOL: int = 1
    DA_MORPH: int = 2
    YO: int = 3
    JYO: int = 4
    SB: int = 5
    COMMON: int = 6
    EOMI: int = 7


class ID(object):
    NONE: int = 0
    PREV: int = 1 << 0
    CONT: int = 1 << 1
    NEXT: int = 1 << 2
    NEXT1: int = 1 << 3
    NEXT2: int = 1 << 4


yo = {
    "가",
    "감",
    "개",
    "걔",
    "걘",
    "괴",
    "까",
    "깨",
    "껴",
    "꿈",
    "꿔",
    "꿰",
    "끔",
    "낌",
    "나",
    "남",
    "내",
    "냄",
    "놂",
    "놔",
    "눠",
    "늚",
    "대",
    "댐",
    "데",
    "돼",
    "되",
    "됨",
    "둠",
    "듦",
    "듬",
    "따",
    "땀",
    "떠",
    "떼",
    "뜀",
    "뜸",
    "띔",
    "매",
    "메",
    "배",
    "베",
    "봬",
    "뵈",
    "빎",
    "빔",
    "빪",
    "뺌",
    "삠",
    "사",
    "삶",
    "삼",
    "새",
    "서",
    "세",
    "셈",
    "싸",
    "쌈",
    "쌔",
    "써",
    "썲",
    "쎄",
    "쏨",
    "쏴",
    "쓺",
    "씀",
    "앎",
    "암",
    "얘",
    "얜",
    "엶",
    "옮",
    "옴",
    "와",
    "왜",
    "자",
    "재",
    "잼",
    "쟤",
    "쟨",
    "젊",
    "져",
    "졺",
    "줌",
    "줘",
    "짜",
    "째",
    "쪄",
    "쬠",
    "찜",
    "차",
    "채",
    "쳐",
    "춤",
    "춰",
    "캐",
    "커",
    "켜",
    "켬",
    "킴",
    "타",
    "탐",
    "터",
    "텨",
    "튐",
    "틈",
    "팜",
    "패",
    "퍼",
    "펌",
    "펴",
    "폄",
    "품",
    "핌",
    "함",
    "해",
}

jyo = {
    "가",
    "갉",
    "갔",
    "갖",
    "같",
    "갚",
    "개",
    "걔",
    "걷",
    "걸",
    "검",
    "겪",
    "골",
    "곪",
    "곱",
    "괴",
    "굵",
    "굶",
    "굼",
    "굽",
    "긁",
    "긋",
    "길",
    "깊",
    "깎",
    "깠",
    "깨",
    "깼",
    "꺾",
    "껐",
    "꼈",
    "꼽",
    "꽂",
    "꾸",
    "꿇",
    "꿨",
    "꿰",
    "뀌",
    "끊",
    "끌",
    "끓",
    "끼",
    "낚",
    "날",
    "낡",
    "남",
    "났",
    "낮",
    "내",
    "냈",
    "넓",
    "넘",
    "넣",
    "녹",
    "놀",
    "높",
    "놓",
    "놨",
    "누",
    "눕",
    "늙",
    "늦",
    "닦",
    "닫",
    "달",
    "닮",
    "닳",
    "답",
    "닿",
    "대",
    "댔",
    "덜",
    "덥",
    "덮",
    "데",
    "뎄",
    "돋",
    "돌",
    "돕",
    "돼",
    "됐",
    "되",
    "두",
    "뒀",
    "듣",
    "들",
    "딛",
    "딪",
    "따",
    "땄",
    "땋",
    "땠",
    "떨",
    "떴",
    "떼",
    "뛰",
    "뜨",
    "뜯",
    "띄",
    "띠",
    "막",
    "많",
    "말",
    "맑",
    "맞",
    "맡",
    "매",
    "맵",
    "맸",
    "맺",
    "먹",
    "멀",
    "메",
    "멨",
    "몰",
    "묵",
    "묶",
    "묻",
    "묽",
    "뭍",
    "믿",
    "밀",
    "밉",
    "박",
    "받",
    "밝",
    "밟",
    "배",
    "뱄",
    "뱉",
    "벌",
    "벗",
    "베",
    "보",
    "볶",
    "봤",
    "봬",
    "뵀",
    "뵈",
    "붇",
    "불",
    "붉",
    "붓",
    "붙",
    "비",
    "빌",
    "빚",
    "빨",
    "빻",
    "빼",
    "뺐",
    "뻗",
    "뻤",
    "뼜",
    "사",
    "살",
    "삵",
    "샀",
    "새",
    "샌",
    "샛",
    "샜",
    "서",
    "섞",
    "섰",
    "세",
    "셌",
    "속",
    "솎",
    "솟",
    "숨",
    "쉬",
    "쉽",
    "시",
    "식",
    "싣",
    "싫",
    "싶",
    "싸",
    "쌌",
    "쌓",
    "쌔",
    "쌨",
    "써",
    "썩",
    "썰",
    "썼",
    "쎄",
    "쏘",
    "쏟",
    "쏴",
    "쐈",
    "쑤",
    "쓰",
    "쓸",
    "씌",
    "씹",
    "앉",
    "않",
    "알",
    "앓",
    "약",
    "얇",
    "얕",
    "얘",
    "얹",
    "얻",
    "얼",
    "없",
    "엎",
    "엮",
    "열",
    "옅",
    "옛",
    "오",
    "온",
    "옭",
    "옮",
    "옳",
    "와",
    "왔",
    "울",
    "읊",
    "일",
    "읽",
    "잃",
    "입",
    "있",
    "잊",
    "자",
    "작",
    "잡",
    "잤",
    "잦",
    "재",
    "잰",
    "쟀",
    "쟤",
    "적",
    "절",
    "젊",
    "접",
    "젓",
    "졌",
    "졸",
    "좁",
    "좇",
    "좋",
    "주",
    "죽",
    "줍",
    "줬",
    "쥐",
    "지",
    "질",
    "집",
    "짓",
    "짖",
    "짙",
    "짜",
    "짧",
    "짰",
    "째",
    "쨌",
    "쩔",
    "쪘",
    "쬐",
    "찌",
    "찍",
    "찐",
    "찝",
    "찢",
    "차",
    "찼",
    "찾",
    "채",
    "챘",
    "쳐",
    "쳤",
    "추",
    "춥",
    "춰",
    "췄",
    "치",
    "캐",
    "캤",
    "커",
    "컸",
    "켜",
    "켠",
    "켰",
    "크",
    "키",
    "타",
    "탔",
    "터",
    "튀",
    "트",
    "파",
    "팔",
    "팠",
    "패",
    "팼",
    "펐",
    "펴",
    "폈",
    "피",
    "하",
    "핥",
    "했",
    "휘",
    "희",
}

ham = {
    "리",
    "절",
    "용",
    "편",
    "륭",
    "듯",
    "야",
    "족",
    "못",
    "끗",
    "안",
    "천",
    "정",
    "각",
    "실",
    "소",
    "끔",
    "분",
    "이",
    "약",
}

um = {
    "았",
    "었",
    "했",
    "없",
    "좋",
    "있",
    "웠",
    "였",
    "않",
    "같",
    "많",
    "적",
    "겠",
    "찮",
    "났",
    "좁",
    "작",
    "싶",
    "셨",
    "졌",
    "좁",
    "넓",
}


before = {
    # 조사
    "이",
    "가",
    "에서",
    "은",
    "는",
    "을",
    "를",
    "도",
    "에",
    "게",
    "께",
    "한테",
    "로",
    "써",
    "와",
    "과",
    "랑",
    "까지",
    "부터",
    "뿐",
    "만",
    "따라",
    "토록",
    "도록",
    "든지",
    "던지",
    "란",
    "만큼",
    "만치",
    "때",
    # 부사
    "너무",
    "잘",
    "못",
    "빨리",
    "매우",
    "몹시",
    "별로",
    "아까",
    "내일",
    "일찍",
    "금방",
    "이미",
    "이리",
    "저리",
    "아니",
    "과연",
    "설마",
    "제발",
    "정말",
    "결코",
    "가득",
    "히",
    # 대명사
    "나",
    "저",
    "우리",
    "저희",
    "너",
    "너희",
    "당신",
    "그대",
    "그",
    "그녀",
    "분",
    "놈",
    "거",
    "것",
    "여기",
    "저기",
    "쪽",
    "곳",
    "님",
}


def create_dict(d, default: Any = 0):
    return defaultdict(lambda: default, d)


Table = create_dict(
    {
        Stats.DA_EOJEOL:
        # EC, EF 주의 !!
        create_dict(
            {
                "갔": ID.PREV,
                "간": ID.PREV,
                "겠": ID.PREV,
                "겼": ID.PREV,
                "같": ID.PREV,
                "놨": ID.PREV,
                "녔": ID.PREV,
                "니": ID.PREV,
                "논": ID.PREV,
                "낸": ID.PREV,
                "냈": ID.PREV,
                "뒀": ID.PREV,
                "때": ID.PREV,
                "랐": ID.PREV,
                "럽": ID.PREV,
                "렵": ID.PREV,
                "렸": ID.PREV,
                "뤘": ID.PREV,
                "몄": ID.PREV,
                "밌": ID.PREV,
                "볐": ID.PREV,
                "볍": ID.PREV,
                "봤": ID.PREV,
                "섰": ID.PREV,
                "샜": ID.PREV,
                "셨": ID.PREV,
                "싼": ID.PREV,
                "싸": ID.PREV,
                "않": ID.PREV,
                "았": ID.PREV,
                "없": ID.PREV,
                "었": ID.PREV,
                "였": ID.PREV,
                "온": ID.PREV,
                "웠": ID.PREV,
                "이": ID.PREV,
                "인": ID.PREV,
                "있": ID.PREV,
                "진": ID.PREV,
                "졌": ID.PREV,
                "쳤": ID.PREV,
                "췄": ID.PREV,
                "챘": ID.PREV,
                "켰": ID.PREV,
                "켠": ID.PREV,
                "팠": ID.PREV,
                "펐": ID.PREV,
                "폈": ID.PREV,
                "했": ID.PREV,
                "혔": ID.PREV,
                "한": ID.NEXT,
                "가": ID.NEXT,
                "고": ID.NEXT | ID.NEXT2,
                "는": ID.NEXT | ID.NEXT2,
                "라": ID.NEXT,
                "시": ID.NEXT,
                "등": ID.NEXT,
                "던": ID.NEXT,
                "든": ID.NEXT,
                "지": ID.NEXT2,
                "를": ID.NEXT,
                "운": ID.NEXT,  # ~ 다운
                "만": ID.NEXT,  # ~ 하다만
                "며": ID.NEXT | ID.NEXT2,
                "면": ID.NEXT | ID.NEXT1 | ID.NEXT2,
                "서": ID.NEXT2,
                "싶": ID.PREV | ID.NEXT,
                "죠": ID.NEXT,
                "죵": ID.NEXT,
                "쥬": ID.NEXT,
                "하": ID.NEXT1,
                "해": ID.NEXT1,
                "도": ID.NEXT2,
                "": ID.NONE,
            }
        ),
        Stats.DA_MORPH:
        # EC, EF 주의 !!
        create_dict(
            {
                "간": ID.PREV,
                "갈": ID.PREV,
                "갉": ID.PREV,
                "감": ID.PREV,
                "갔": ID.PREV,
                "같": ID.PREV,
                "갚": ID.PREV,
                "개": ID.PREV,
                "걔": ID.PREV,
                "건": ID.PREV,
                "검": ID.PREV,
                "겪": ID.PREV,
                "곤": ID.PREV,
                "골": ID.PREV,
                "곪": ID.PREV,
                "곱": ID.PREV,
                "괴": ID.PREV,
                "군": ID.PREV,
                "굵": ID.PREV,
                "굶": ID.PREV,
                "굽": ID.PREV,
                "긁": ID.PREV,
                "긋": ID.PREV,
                "길": ID.PREV,
                "깊": ID.PREV,
                "까": ID.PREV,
                "깎": ID.PREV,
                "깐": ID.PREV,
                "깠": ID.PREV,
                "깨": ID.PREV,
                "깬": ID.PREV,
                "깼": ID.PREV,
                "꺾": ID.PREV,
                "껐": ID.PREV,
                "꼈": ID.PREV,
                "꼬": ID.PREV,
                "꼽": ID.PREV,
                "꽂": ID.PREV,
                "꾸": ID.PREV,
                "꾼": ID.PREV,
                "꿨": ID.PREV,
                "꿰": ID.PREV,
                "끼": ID.PREV,
                "낀": ID.PREV,
                "나": ID.PREV,
                "낚": ID.PREV,
                "난": ID.PREV,
                "날": ID.PREV,
                "낡": ID.PREV,
                "남": ID.PREV,
                "났": ID.PREV,
                "낮": ID.PREV,
                "내": ID.PREV,
                "낸": ID.PREV,
                "냈": ID.PREV,
                "넓": ID.PREV,
                "넘": ID.PREV,
                "넣": ID.PREV,
                "녹": ID.PREV,
                "논": ID.PREV,
                "놀": ID.PREV,
                "높": ID.PREV,
                "놓": ID.PREV,
                "놨": ID.PREV,
                "누": ID.PREV,
                "눈": ID.PREV,
                "눕": ID.PREV,
                "늘": ID.PREV,
                "늙": ID.PREV,
                "늦": ID.PREV,
                "닦": ID.PREV,
                "단": ID.PREV,
                "닫": ID.PREV,
                "달": ID.PREV,
                "닮": ID.PREV,
                "닳": ID.PREV,
                "담": ID.PREV,
                "답": ID.PREV,
                "닿": ID.PREV,
                "대": ID.PREV,
                "댄": ID.PREV,
                "댔": ID.PREV,
                "덜": ID.PREV,
                "덥": ID.PREV,
                "덮": ID.PREV,
                "데": ID.PREV,
                "덴": ID.PREV,
                "뎄": ID.PREV,
                "돈": ID.PREV,
                "돋": ID.PREV,
                "돌": ID.PREV,
                "돕": ID.PREV,
                "돼": ID.PREV,
                "됐": ID.PREV,
                "되": ID.PREV,
                "된": ID.PREV,
                "두": ID.PREV,
                "둔": ID.PREV,
                "둠": ID.PREV,
                "뒀": ID.PREV,
                "듣": ID.PREV,
                "들": ID.PREV,
                "딛": ID.PREV,
                "딪": ID.PREV,
                "따": ID.PREV,
                "딴": ID.PREV,
                "땄": ID.PREV,
                "땋": ID.PREV,
                "땠": ID.PREV,
                "떨": ID.PREV,
                "떴": ID.PREV,
                "떼": ID.PREV,
                "뗀": ID.PREV,
                "뛰": ID.PREV,
                "뜨": ID.PREV,
                "뜯": ID.PREV,
                "띄": ID.PREV,
                "띈": ID.PREV,
                "띠": ID.PREV,
                "막": ID.PREV,
                "많": ID.PREV,
                "말": ID.PREV,
                "맑": ID.PREV,
                "맞": ID.PREV,
                "맡": ID.PREV,
                "매": ID.PREV,
                "맨": ID.PREV,
                "맵": ID.PREV,
                "맸": ID.PREV,
                "맺": ID.PREV,
                "먹": ID.PREV,
                "멀": ID.PREV,
                "메": ID.PREV,
                "멘": ID.PREV,
                "멨": ID.PREV,
                "몬": ID.PREV,
                "몰": ID.PREV,
                "묵": ID.PREV,
                "묶": ID.PREV,
                "묻": ID.PREV,
                "물": ID.PREV,
                "묽": ID.PREV,
                "뭍": ID.PREV,
                "뭘": ID.PREV,
                "민": ID.PREV,
                "믿": ID.PREV,
                "밀": ID.PREV,
                "밉": ID.PREV,
                "박": ID.PREV,
                "받": ID.PREV,
                "밝": ID.PREV,
                "밟": ID.PREV,
                "배": ID.PREV,
                "밴": ID.PREV,
                "뱄": ID.PREV,
                "뱉": ID.PREV,
                "번": ID.PREV,
                "벌": ID.PREV,
                "벗": ID.PREV,
                "베": ID.PREV,
                "벤": ID.PREV,
                "보": ID.PREV,
                "볶": ID.PREV,
                "본": ID.PREV,
                "봤": ID.PREV,
                "봬": ID.PREV,
                "뵀": ID.PREV,
                "뵈": ID.PREV,
                "뵌": ID.PREV,
                "분": ID.PREV,
                "붇": ID.PREV,
                "불": ID.PREV,
                "붉": ID.PREV,
                "붓": ID.PREV,
                "붙": ID.PREV,
                "비": ID.PREV,
                "빈": ID.PREV,
                "빌": ID.PREV,
                "빚": ID.PREV,
                "빤": ID.PREV,
                "빨": ID.PREV,
                "빻": ID.PREV,
                "빼": ID.PREV,
                "뺀": ID.PREV,
                "뺐": ID.PREV,
                "뻗": ID.PREV,
                "뻤": ID.PREV,
                "뼜": ID.PREV,
                "삔": ID.PREV,
                "사": ID.PREV,
                "산": ID.PREV,
                "살": ID.PREV,
                "삵": ID.PREV,
                "샀": ID.PREV,
                "새": ID.PREV,
                "샌": ID.PREV,
                "샜": ID.PREV,
                "섞": ID.PREV,
                "선": ID.PREV,
                "섰": ID.PREV,
                "세": ID.PREV,
                "셌": ID.PREV,
                "속": ID.PREV,
                "솎": ID.PREV,
                "솟": ID.PREV,
                "숨": ID.PREV,
                "수": ID.PREV,  # ~ 했수다.
                "쉬": ID.PREV,
                "쉰": ID.PREV,
                "쉽": ID.PREV,
                "식": ID.PREV,
                "싣": ID.PREV,
                "싫": ID.PREV,
                "싸": ID.PREV,
                "싼": ID.PREV,
                "쌌": ID.PREV,
                "쌓": ID.PREV,
                "쌔": ID.PREV,
                "쌨": ID.PREV,
                "썩": ID.PREV,
                "썰": ID.PREV,
                "썼": ID.PREV,
                "쎄": ID.PREV,
                "쏘": ID.PREV,
                "쏜": ID.PREV,
                "쏟": ID.PREV,
                "쐈": ID.PREV,
                "쓰": ID.PREV,
                "쓴": ID.PREV,
                "쓸": ID.PREV,
                "씹": ID.PREV,
                "안": ID.PREV,
                "앉": ID.PREV,
                "않": ID.PREV,
                "알": ID.PREV,
                "앓": ID.PREV,
                "약": ID.PREV,
                "얇": ID.PREV,
                "얕": ID.PREV,
                "얘": ID.PREV,
                "언": ID.PREV,
                "얹": ID.PREV,
                "얻": ID.PREV,
                "얼": ID.PREV,
                "없": ID.PREV,
                "엎": ID.PREV,
                "엮": ID.PREV,
                "연": ID.PREV,
                "열": ID.PREV,
                "옅": ID.PREV,
                "오": ID.PREV,
                "온": ID.PREV,
                "옭": ID.PREV,
                "옳": ID.PREV,
                "왔": ID.PREV,
                "울": ID.PREV,
                "읊": ID.PREV,
                "일": ID.PREV,
                "읽": ID.PREV,
                "잃": ID.PREV,
                "입": ID.PREV,
                "있": ID.PREV,
                "잊": ID.PREV,
                "자": ID.PREV,
                "작": ID.PREV,
                "잔": ID.PREV,
                "잡": ID.PREV,
                "잤": ID.PREV,
                "잦": ID.PREV,
                "재": ID.PREV,
                "잰": ID.PREV,
                "쟀": ID.PREV,
                "쟤": ID.PREV,
                "적": ID.PREV,
                "전": ID.PREV,
                "절": ID.PREV,
                "젊": ID.PREV,
                "접": ID.PREV,
                "젓": ID.PREV,
                "졌": ID.PREV,
                "존": ID.PREV,
                "졸": ID.PREV,
                "좁": ID.PREV,
                "좋": ID.PREV,
                "주": ID.PREV,
                "죽": ID.PREV,
                "준": ID.PREV,
                "줍": ID.PREV,
                "줬": ID.PREV,
                "쥐": ID.PREV,
                "진": ID.PREV,
                "질": ID.PREV,
                "집": ID.PREV,
                "짓": ID.PREV,
                "짖": ID.PREV,
                "짙": ID.PREV,
                "짜": ID.PREV,
                "짧": ID.PREV,
                "짰": ID.PREV,
                "째": ID.PREV,
                "짼": ID.PREV,
                "쨌": ID.PREV,
                "쩐": ID.PREV,
                "쩔": ID.PREV,
                "쪘": ID.PREV,
                "쫀": ID.PREV,
                "쬐": ID.PREV,
                "찌": ID.PREV,
                "찍": ID.PREV,
                "찐": ID.PREV,
                "찝": ID.PREV,
                "찢": ID.PREV,
                "차": ID.PREV,
                "찬": ID.PREV,
                "참": ID.PREV,
                "찼": ID.PREV,
                "찾": ID.PREV,
                "채": ID.PREV,
                "챈": ID.PREV,
                "챘": ID.PREV,
                "쳤": ID.PREV,
                "추": ID.PREV,
                "춘": ID.PREV,
                "춥": ID.PREV,
                "췄": ID.PREV,
                "치": ID.PREV,
                "친": ID.PREV,
                "캐": ID.PREV,
                "캤": ID.PREV,
                "컸": ID.PREV,
                "켜": ID.PREV,
                "켠": ID.PREV,
                "켰": ID.PREV,
                "크": ID.PREV,
                "키": ID.PREV,
                "킨": ID.PREV,
                "타": ID.PREV,
                "탄": ID.PREV,
                "탔": ID.PREV,
                "튀": ID.PREV,
                "튄": ID.PREV,
                "트": ID.PREV,
                "튼": ID.PREV,
                "파": ID.PREV,
                "팔": ID.PREV,
                "팠": ID.PREV,
                "패": ID.PREV,
                "팼": ID.PREV,
                "펐": ID.PREV,
                "펴": ID.PREV,
                "편": ID.PREV,
                "폈": ID.PREV,
                "푼": ID.PREV,
                "품": ID.PREV,
                "피": ID.PREV,
                "핀": ID.PREV,
                "핥": ID.PREV,
                "했": ID.PREV,
                "헌": ID.PREV,
                "휘": ID.PREV,
                "희": ID.PREV,
                "겠": ID.PREV,
                "겼": ID.PREV,
                "녔": ID.PREV,
                "니": ID.PREV,
                "때": ID.PREV,
                "랐": ID.PREV,
                "럽": ID.PREV,
                "렵": ID.PREV,
                "렸": ID.PREV,
                "뤘": ID.PREV,
                "몄": ID.PREV,
                "밌": ID.PREV,
                "볐": ID.PREV,
                "볍": ID.PREV,
                "셨": ID.PREV,
                "았": ID.PREV,
                "었": ID.PREV,
                "였": ID.PREV,
                "웠": ID.PREV,
                "이": ID.PREV,
                "인": ID.PREV,
                "혔": ID.PREV,
                "한": ID.NEXT,
                "가": ID.PREV | ID.NEXT,
                "고": ID.NEXT | ID.NEXT2,
                "구": ID.NEXT | ID.NEXT2,
                "는": ID.NEXT | ID.NEXT2,
                "라": ID.NEXT,
                "시": ID.PREV | ID.NEXT,
                "등": ID.NEXT,
                "던": ID.PREV | ID.NEXT,
                "든": ID.PREV | ID.NEXT,
                "지": ID.PREV | ID.NEXT2,
                "를": ID.NEXT,
                "운": ID.PREV | ID.NEXT,  # ~ 다운
                "만": ID.NEXT,
                "며": ID.NEXT | ID.NEXT2,
                "면": ID.NEXT | ID.NEXT1 | ID.NEXT2,
                "서": ID.PREV | ID.NEXT2,
                "싶": ID.PREV | ID.NEXT,
                "죠": ID.NEXT,
                "죵": ID.NEXT,
                "쥬": ID.NEXT,
                "하": ID.PREV | ID.NEXT1,
                "거": ID.PREV | ID.NEXT,
                "해": ID.NEXT1,
                "도": ID.NEXT2,
                "": ID.NONE,
            }
        ),
        Stats.YO: create_dict(
            {
                "겨": ID.PREV,
                "거": ID.PREV,
                "구": ID.PREV,
                "군": ID.PREV,
                "걸": ID.PREV,
                "까": ID.PREV,
                "께": ID.PREV,
                "껴": ID.PREV,
                "네": ID.PREV,
                "나": ID.PREV,
                "니": ID.PREV,
                "데": ID.PREV,
                "든": ID.PREV,
                "려": ID.PREV,
                "서": ID.PREV,
                "세": ID.PREV,
                "아": ID.PREV,
                "어": ID.PREV,
                "워": ID.PREV,
                "에": ID.PREV,
                "예": ID.PREV,
                "을": ID.PREV,
                "져": ID.PREV,
                "줘": ID.PREV,
                "지": ID.PREV,
                "춰": ID.PREV,
                "해": ID.PREV,
                "먼": ID.PREV,
                "만": ID.PREV,
                "고": ID.NEXT2,
                "는": ID.NEXT,
                "라": ID.NEXT1,
                "등": ID.NEXT,
                "를": ID.NEXT,
                "즘": ID.NEXT,
                "소": ID.NEXT,
                "며": ID.NEXT2,
                "면": ID.PREV | ID.NEXT2,
                "하": ID.NEXT1,
                "": ID.NONE,
            }
        ),
        Stats.JYO: create_dict(
            {
                "거": ID.PREV,
                "가": ID.PREV,
                "갔": ID.PREV,
                "겠": ID.PREV,
                "같": ID.PREV,
                "놨": ID.PREV,
                "녔": ID.PREV,
                "냈": ID.PREV,
                "니": ID.PREV,
                "뒀": ID.PREV,
                "았": ID.PREV,
                "르": ID.PREV,
                "랐": ID.PREV,
                "럽": ID.PREV,
                "렵": ID.PREV,
                "렸": ID.PREV,
                "맞": ID.PREV,
                "몄": ID.PREV,
                "밌": ID.PREV,
                "볐": ID.PREV,
                "볍": ID.PREV,
                "봤": ID.PREV,
                "서": ID.PREV,
                "섰": ID.PREV,
                "셨": ID.PREV,
                "샜": ID.PREV,
                "않": ID.PREV,
                "없": ID.PREV,
                "었": ID.PREV,
                "였": ID.PREV,
                "이": ID.PREV,
                "졌": ID.PREV,
                "쳤": ID.PREV,
                "챘": ID.PREV,
                "켰": ID.PREV,
                "팠": ID.PREV,
                "폈": ID.PREV,
                "하": ID.PREV,
                "했": ID.PREV,
                "혔": ID.PREV,
                "고": ID.PREV | ID.NEXT2,
                "는": ID.NEXT,
                "등": ID.NEXT,
                "라": ID.NEXT1,
                "를": ID.NEXT,
                "며": ID.NEXT2,
                "면": ID.PREV | ID.NEXT2,
                "": ID.NONE,
            }
        ),
        Stats.SB:
        # https://www.yeoju.go.kr/history/jsp/Theme/Save_View.jsp?BC_ID=d0400
        create_dict(
            {
                "것": ID.PREV,
                "가": ID.PREV,
                "까": ID.PREV,
                "걸": ID.PREV,
                "껄": ID.PREV,
                "나": ID.PREV,
                "니": ID.PREV,
                "네": ID.PREV,
                "다": ID.PREV,
                "쎄": ID.PREV,
                "래": ID.PREV,
                "데": ID.PREV,
                "지": ID.PREV,
                "든": ID.PREV,
                "덩": ID.PREV,
                "등": ID.PREV,
                "랴": ID.PREV,
                "마": ID.PREV,
                "봐": ID.PREV,
                "서": ID.PREV,
                "셈": ID.PREV,  # ~하셈 (신조어)
                "아": ID.PREV,
                "어": ID.PREV,
                "오": ID.PREV,
                "요": ID.PREV,
                "용": ID.PREV,
                "을": ID.PREV,
                "자": ID.PREV,
                "죠": ID.PREV,
                "쥬": ID.PREV,  # ~했쥬 (사투리)
                "죵": ID.PREV,  # ~했죵 (신조어)
                "고": ID.NEXT2,
                "는": ID.NEXT,
                "라": ID.PREV | ID.NEXT,
                "며": ID.NEXT2,
                "면": ID.NEXT2,
                "하": ID.NEXT1,
                "": ID.NONE,
            }
        ),
        Stats.COMMON: create_dict(
            {
                "ㄱ": ID.CONT,
                "ㄴ": ID.CONT,
                "ㄷ": ID.CONT,
                "ㄹ": ID.CONT,
                "ㅁ": ID.CONT,
                "ㅂ": ID.CONT,
                "ㅅ": ID.CONT,
                "ㅇ": ID.CONT,
                "ㅈ": ID.CONT,
                "ㅊ": ID.CONT,
                "ㅋ": ID.CONT,
                "ㅌ": ID.CONT,
                "ㅍ": ID.CONT,
                "ㅎ": ID.CONT,
                "ㅏ": ID.CONT,
                "ㅑ": ID.CONT,
                "ㅓ": ID.CONT,
                "ㅕ": ID.CONT,
                "ㅗ": ID.CONT,
                "ㅛ": ID.CONT,
                "ㅜ": ID.CONT,
                "ㅠ": ID.CONT,
                "ㅡ": ID.CONT,
                "ㅣ": ID.CONT,
                "^": ID.CONT,
                ";": ID.CONT,
                ".": ID.CONT,
                "?": ID.CONT,
                "!": ID.CONT,
                "~": ID.CONT,
                "…": ID.CONT,
                "": ID.NONE,
            }
        ),
        Stats.EOMI: create_dict(
            {
                "나": ID.NEXT,  # 했다나 뭐라나, 한다나 뭐라나
                "고": ID.NEXT | ID.NEXT2,
                "구": ID.NEXT | ID.NEXT2,
                "라": ID.NEXT,
                "시": ID.NEXT,
                "다": ID.NEXT,
                "등": ID.NEXT,
                "던": ID.NEXT,
                "든": ID.NEXT,
                "지": ID.NEXT2,
                "요": ID.NEXT,  # ~하구요
                "유": ID.NEXT,  # ~하구유
                "용": ID.NEXT,  # ~하군용
                "만": ID.NEXT,
                "은": ID.NEXT | ID.NEXT2,
                "는": ID.NEXT | ID.NEXT2,
                "이": ID.NEXT,
                "가": ID.NEXT,
                "을": ID.NEXT,
                "를": ID.NEXT,
                "의": ID.NEXT,
                "며": ID.NEXT | ID.NEXT2,
                "면": ID.NEXT | ID.NEXT1 | ID.NEXT2,
                "서": ID.NEXT2,
                "싶": ID.NEXT,
                "죠": ID.NEXT,
                "죵": ID.NEXT,
                "쥬": ID.NEXT,
                "하": ID.NEXT1,
                "해": ID.NEXT1,
                "도": ID.NEXT | ID.NEXT2,
            }
        ),
    },
    default=create_dict({}),
)

post_processing_da = list(
    set([f"{x} {_}{y} " for y in ["다"] for _ in yo for x in before])
)
post_processing_yo = list(
    set([f"{x} {_}{y} " for y in ["요"] for _ in yo for x in before])
)
post_processing_jyo = list(
    set([f"{x} {_}{j} " for j in ["죠"] for _ in jyo for x in before])
)



# -*- coding:utf-8 -*-
"""두음법칙에 관련된 내용은
http://ko.wikipedia.org/wiki/%EB%91%90%EC%9D%8C_%EB%B2%95%EC%B9%99 를 참고.
"""
from .impl import translate
import warnings


__all__ = ["is_hanja", "is_valid_mode", "split_hanja", "translate"]
__author__ = "Sumin Byeon"
__email__ = "suminb@gmail.com"
__version__ = "0.13.3"


# Copied from https://wiki.python.org/moin/PythonDecoratorLibrary
def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    def new_func(*args, **kwargs):
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
        )
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


def lazily_import(import_string):
    import_path, func_name = import_string.split(":")

    def load_and_call(*args, **kwargs):
        mod = __import__(import_path)
        for mod_name in import_path.split(".")[1:]:
            mod = getattr(mod, mod_name)
        func = getattr(mod, func_name)
        globals()[func_name] = func
        return func(*args, **kwargs)

    globals()[func_name] = load_and_call


# lazily_import("hanja.impl:is_hanja")
# lazily_import("hanja.impl:is_valid_mode")
# lazily_import("hanja.impl:split_hanja")
# lazily_import("hanja.impl:translate")



import re
import collections
import copy
from math import log


class UnigramTokenizer:
    def __init__(
        self,
        initial_vocab_size=2000,
        vocab_size=1000,
        whitespace_token="▁",
        unknown_token="[unk]",
        lowercase=True,
        percent_to_prune=0.1,
    ):
        self.vocab_tokenized = {}

        self.word_freqs = {}
        self.subword_freqs = {}
        self.token_freqs = {}
        self.inital_vocab = {}
        self.vocab = {}

        self.initial_vocab_size = initial_vocab_size
        self.vocab_size = vocab_size
        self.suffix_indicator = whitespace_token
        self.unknown_token = unknown_token
        self.lowercase = lowercase
        self.percent_to_prune = percent_to_prune

    def initialize_vocab(self, texts, initial_vocab_size=None):
        if initial_vocab_size is None:
            initial_vocab_size = self.initial_vocab_size
        self.word_freqs = self.get_word_frequency(texts)
        sorted_subwords, character_freqs = self.initialize_subwords(self.word_freqs)
        token_freqs = (
            list(character_freqs.items())
            + sorted_subwords[: self.initial_vocab_size - len(character_freqs)]
        )
        self.token_freqs = {token: freq for token, freq in token_freqs}

        total_sum = sum([freq for _, freq in self.token_freqs.items()])
        self.inital_vocab = {
            token: -log(freq / total_sum) for token, freq in self.token_freqs.items()
        }

    def get_word_frequency(self, texts):
        word_freqs = {}
        for text in texts:
            words = self.pre_tokenize(text)
            for word in words:
                word_freqs[word] = word_freqs.get(word, 0) + 1
        return word_freqs

    def initialize_subwords(self, word_freqs, verbose=True):
        character_freqs = collections.defaultdict(int)
        subwords_freqs = collections.defaultdict(int)
        for word, freq in word_freqs.items():
            for i in range(len(word)):
                character_freqs[word[i]] += freq
                # Loop through the subwords of length at least 2
                for j in range(i + 2, len(word) + 1):
                    subwords_freqs[word[i:j]] += freq

        # Sort subwords by frequency
        sorted_subwords = sorted(
            subwords_freqs.items(), key=lambda x: x[1], reverse=True
        )
        if verbose:
            print("Top 10 subwords: {}".format(sorted_subwords[:10]))
        return sorted_subwords, character_freqs

    def encode_word(self, word, vocab):
        best_segmentations = [{"start": 0, "score": 1}] + [
            {"start": None, "score": None} for _ in range(len(word))
        ]
        for start_idx in range(len(word)):
            # This should be properly filled by the previous steps of the loop
            best_score_at_start = best_segmentations[start_idx]["score"]
            for end_idx in range(start_idx + 1, len(word) + 1):
                token = word[start_idx:end_idx]
                if token in vocab and best_score_at_start is not None:
                    score = vocab[token] + best_score_at_start
                    # If we have found a better segmentation ending at end_idx, we update
                    if (
                        best_segmentations[end_idx]["score"] is None
                        or best_segmentations[end_idx]["score"] > score
                    ):
                        best_segmentations[end_idx] = {
                            "start": start_idx,
                            "score": score,
                        }

        segmentation = best_segmentations[-1]
        if segmentation["score"] is None:
            # We did not find a tokenization of the word -> unknown
            return [self.unknown_token], None

        score = segmentation["score"]
        start = segmentation["start"]
        end = len(word)
        tokens = []
        while start != 0:
            tokens.insert(0, word[start:end])
            next_start = best_segmentations[start]["start"]
            end = start
            start = next_start
        tokens.insert(0, word[start:end])
        return tokens, score

    def compute_loss(self, vocab):
        loss = 0
        for word, freq in self.word_freqs.items():
            _, word_loss = self.encode_word(word, vocab)
            loss += freq * word_loss
        return loss

    def compute_scores(self, vocab):
        scores = {}
        vocab_loss = self.compute_loss(vocab)
        for token, _ in vocab.items():
            # We always keep tokens of length 1
            if len(token) == 1:
                continue
            vocab_without_token = copy.deepcopy(vocab)
            _ = vocab_without_token.pop(token)
            scores[token] = self.compute_loss(vocab_without_token) - vocab_loss
        return scores

    def fit(
        self,
        texts,
        vocab_size=None,
        percent_to_prune=None,
        verbose=True,
        print_every=2,
    ):
        if vocab_size is None:
            vocab_size = self.vocab_size
        if percent_to_prune is None:
            percent_to_prune = self.percent_to_prune

        self.initialize_vocab(texts)
        vocab = copy.deepcopy(self.inital_vocab)
        iter = 0

        while len(vocab) > vocab_size:
            scores = self.compute_scores(vocab)
            sorted_scores = sorted(scores.items(), key=lambda x: x[1])
            # Remove percent_to_remove tokens with the lowest scores.
            for i in range(int(len(vocab) * percent_to_prune)):
                if len(vocab) <= vocab_size:
                    break
                _ = self.token_freqs.pop(sorted_scores[i][0])

            total_sum = sum([freq for _, freq in self.token_freqs.items()])
            vocab = {
                token: -log(freq / total_sum)
                for token, freq in self.token_freqs.items()
            }

            if verbose and print_every > 0 and i % print_every == 0:
                print("Iteration: {}, vocab size: {}".format(iter, len(vocab)))
            iter += 1

        self.vocab = vocab
        if verbose:
            print("Final vocab size: {}".format(len(self.vocab)))

    def pre_tokenize(self, text):
        if self.lowercase:
            text = text.lower()
        text = re.sub(r"\s+", " ", text)
        return text.split()

    def tokenize(self, text):
        words = self.pre_tokenize(text)
        encoded_words = [self.encode_word(word, self.vocab)[0] for word in words]
        return sum(encoded_words, [])



from typing import Any, Callable, List, Mapping, Optional

from ...preprocess import BasePreprocessor
from ...types import DataPoint


class LabelingFunction:
    """Base class for labeling functions.

    A labeling function (LF) is a function that takes a data point
    as input and produces an integer label, corresponding to a
    class. A labeling function can also abstain from voting by
    outputting ``-1``. For examples, see the Snorkel tutorials.

    This class wraps a Python function outputting a label. Extra
    functionality, such as running preprocessors and storing
    resources, is provided. Simple LFs can be defined via a
    decorator. See ``labeling_function``.

    Parameters
    ----------
    name
        Name of the LF
    f
        Function that implements the core LF logic
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run on data points before LF execution

    Raises
    ------
    ValueError
        Calling incorrectly defined preprocessors

    Attributes
    ----------
    name
        See above
    """

    def __init__(
        self,
        name: str,
        f: Callable[..., int],
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
    ) -> None:
        self.name = name
        self._f = f
        self._resources = resources or {}
        self._pre = pre or []

    def _preprocess_data_point(self, x: DataPoint) -> DataPoint:
        for preprocessor in self._pre:
            x = preprocessor(x)
            if x is None:
                raise ValueError("Preprocessor should not return None")
        return x

    def __call__(self, x: DataPoint) -> int:
        """Label data point.

        Runs all preprocessors, then passes preprocessed data point to LF.

        Parameters
        ----------
        x
            Data point to label

        Returns
        -------
        int
            Label for data point
        """
        x = self._preprocess_data_point(x)
        return self._f(x, **self._resources)

    def __repr__(self) -> str:
        preprocessor_str = f", Preprocessors: {self._pre}"
        return f"{type(self).__name__} {self.name}{preprocessor_str}"


class labeling_function:
    """Decorator to define a LabelingFunction object from a function.

    Parameters
    ----------
    name
        Name of the LF
    resources
        Labeling resources passed in to ``f`` via ``kwargs``
    pre
        Preprocessors to run on data points before LF execution

    Examples
    --------
    >>> @labeling_function()
    ... def f(x):
    ...     return 0 if x.a > 42 else -1
    >>> f
    LabelingFunction f, Preprocessors: []
    >>> from types import SimpleNamespace
    >>> x = SimpleNamespace(a=90, b=12)
    >>> f(x)
    0

    >>> @labeling_function(name="my_lf")
    ... def g(x):
    ...     return 0 if x.a > 42 else -1
    >>> g
    LabelingFunction my_lf, Preprocessors: []
    """

    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[Mapping[str, Any]] = None,
        pre: Optional[List[BasePreprocessor]] = None,
    ) -> None:
        if callable(name):
            raise ValueError("Looks like this decorator is missing parentheses!")
        self.name = name
        self.resources = resources
        self.pre = pre

    def __call__(self, f: Callable[..., int]) -> LabelingFunction:
        """Wrap a function to create a ``LabelingFunction``.

        Parameters
        ----------
        f
            Function that implements the core LF logic

        Returns
        -------
        LabelingFunction
            New ``LabelingFunction`` executing logic in wrapped function
        """
        name = self.name or f.__name__
        return LabelingFunction(name=name, f=f, resources=self.resources, pre=self.pre)



import logging
import rubrix as rb
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Rubrix:
    def __init__(self, **args):
        args = eKonf.to_config(args)
        self.args = args
        self.name = args.name
        self.verbose = args.get("verbose", False)
        self._api_key = args.get("api_key") or eKonf.osenv("RUBRIX_API_KEY")
        self._api_url = args.get("api_url") or eKonf.osenv("RUBRIX_API_URL")
        self._workspace = args.get("workspace") or eKonf.osenv("RUBRIX_WORKSPACE")
        self.auto = args.auto

        if self.auto.init:
            self.init_rubrix()

    def init_rubrix(self):
        log.info(f"Initializing Rubrix: {self._api_url}, workspace: {self._workspace}")
        rb.init(api_url=self._api_url, api_key=self._api_key, workspace=self._workspace)

    def set_workspace(self, workspace):
        rb.set_workspace(workspace)
        self._workspace = workspace

    def get_workspace(self):
        return rb.get_workspace()

    def create_records_from_cv_preds(
        self,
        cv_preds,
        columns=None,
        meta_columns=["id", "split"],
        prediction_agent=None,
    ):
        columns = columns or self.args.columns
        text_col = columns["text"]
        label_col = columns["labels"]
        pred_col = columns["model_outputs"]
        records = []
        for i, row in cv_preds.iterrows():
            record = rb.TextClassificationRecord(
                inputs={"text": row[text_col]},
                prediction=list(row[pred_col].items()),
                annotation=row[label_col] if label_col in cv_preds.columns else None,
                metadata=row[meta_columns].to_dict(),
                prediction_agent=prediction_agent,
            )
            records.append(record)
        log.info(f"Created {len(records)} records")
        return records

    def find_label_errors(self, records):
        from rubrix.labeling.text_classification import find_label_errors

        label_errors = find_label_errors(records)
        log.info(f"Found {len(label_errors)} label errors")
        return label_errors

    def log(self, records, name):
        rb.log(records, name=name)
        log.info(f"Logged {len(records)} records")

    def delete(self, name):
        rb.delete(name)
        log.info(f"Deleted {name}")

    def copy(self, dataset, name_of_copy):
        rb.copy(dataset, name_of_copy)
        log.info(f"Copied {dataset} to {name_of_copy}")

    def rename(self, dataset, new_name):
        rb.copy(dataset, new_name)
        rb.delete(dataset)
        log.info(f"Renamed {dataset} to {new_name}")

    def load(
        self, name, query=None, ids=None, limit=None, id_from=None, as_pandas=True
    ):
        dset = rb.load(name, query=query, ids=ids, limit=limit, id_from=id_from)
        if as_pandas:
            dset = dset.to_pandas()
        return dset

    def update_label_errors(self, original_data, correct_data, id="id", split="train"):
        log.info(f"Updating {len(correct_data)} records")
        if "annotation_agent" not in original_data.columns:
            original_data["annotation_agent"] = None
        original_data["original_labels"] = original_data["labels"]
        for i, row in correct_data.iterrows():
            correct_labels = row["annotation"]
            agent = row["annotation_agent"]
            metadata = row["metadata"]
            status = row["status"]
            if split and metadata["split"] != split:
                continue
            if status == "Validated":
                original_data.loc[
                    (original_data[id] == metadata[id])
                    & (original_data["split"] == split),
                    ["labels", "annotation_agent", "status"],
                ] = [correct_labels, agent, status]
            elif status == "Discarded":
                original_data = original_data[
                    ~(
                        (original_data[id] == metadata[id])
                        & (original_data["split"] == split)
                    )
                ]
        return original_data

    def remove_label_errors(self, original_data, error_records, id="id", split="train"):
        log.info(f"Removing {len(error_records)} records")
        for record in error_records:
            metadata = record.metadata
            if split and metadata["split"] != split:
                continue
            original_data = original_data[
                ~(
                    (original_data[id] == metadata[id])
                    & (original_data["split"] == split)
                )
            ]

        return original_data



import logging
from functools import partial
from typing import Callable, List, Mapping, Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from torch import nn

from ..analysis import Scorer

Outputs = Mapping[str, List[torch.FloatTensor]]


class Operation:
    """A single operation (forward pass of a module) to execute in a Task.

    See ``Task`` for more detail on the usage and semantics of an Operation.

    Parameters
    ----------
    name
        The name of this operation (defaults to module_name since for most workflows,
        each module is only used once per forward pass)
    module_name
        The name of the module in the module pool that this operation uses
    inputs
        The inputs that the specified module expects, given as a list of names of
        previous operations (or optionally a tuple of the operation name and a key
        if the output of that module is a dict instead of a Tensor).
        Note that the original input to the model can be referred to as "_input_".

    Example
    -------
    >>> op1 = Operation(module_name="linear1", inputs=[("_input_", "features")])
    >>> op2 = Operation(module_name="linear2", inputs=["linear1"])
    >>> op_sequence = [op1, op2]

    Attributes
    ----------
    name
        See above
    module_name
        See above
    inputs
        See above
    """

    def __init__(
        self,
        module_name: str,
        inputs: Sequence[Union[str, Tuple[str, str]]],
        name: Optional[str] = None,
    ) -> None:
        self.name = name or module_name
        self.module_name = module_name
        self.inputs = inputs

    def __repr__(self) -> str:
        return (
            f"Operation(name={self.name}, "
            f"module_name={self.module_name}, "
            f"inputs={self.inputs})"
        )


class Task:
    r"""A single task (a collection of modules and specified path through them).

    Parameters
    ----------
    name
        The name of the task
    module_pool
        A ModuleDict mapping module names to the modules themselves
    op_sequence
        A list of ``Operation``\s to execute in order, defining the flow of information
        through the network for this task
    scorer
        A ``Scorer`` with the desired metrics to calculate for this task
    loss_func
        A function that converts final logits into loss values.
        Defaults to F.cross_entropy() if none is provided.
        To use probalistic labels for training, use the Snorkel-defined method
        cross_entropy_with_probs() instead.
    output_func
        A function that converts final logits into 'outputs' (e.g. probabilities)
        Defaults to F.softmax(..., dim=1).

    Attributes
    ----------
    name
        See above
    module_pool
        See above
    op_sequence
        See above
    scorer
        See above
    loss_func
        See above
    output_func
        See above
    """

    def __init__(
        self,
        name: str,
        module_pool: nn.ModuleDict,
        op_sequence: Sequence[Operation],
        scorer: Scorer = Scorer(metrics=["accuracy"]),
        loss_func: Optional[Callable[..., torch.Tensor]] = None,
        output_func: Optional[Callable[..., torch.Tensor]] = None,
    ) -> None:
        self.name = name
        self.module_pool = module_pool
        self.op_sequence = op_sequence
        self.loss_func = loss_func or F.cross_entropy
        self.output_func = output_func or partial(F.softmax, dim=1)
        self.scorer = scorer

        logging.info(f"Created task: {self.name}")

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"



from pathlib import Path
from omegaconf import SCMode, DictConfig, ListConfig
from typing import Any, List, IO, Dict, Union, Tuple
from ekorpkit.utils.notebook import (
    _clear_output,
    _cprint,
    _create_button,
    _create_dropdown,
    _create_floatslider,
    _create_image,
    _create_radiobutton,
    _create_textarea,
    _display,
    _display_image,
    _get_display,
    _hide_code_in_slideshow,
    _is_notebook,
    _is_colab,
    _read,
)
from .base import (
    __ekorpkit_path__,
    __home_path__,
    __version__,
    _apply,
    _compose,
    _config,
    _Defaults,
    _dependencies,
    _dict_product,
    _dict_to_dataframe,
    _ensure_kwargs,
    _ensure_list,
    _env_set,
    _exists,
    _function,
    _getLogger,
    _getsource,
    _init_env_,
    _instantiate,
    _is_config,
    _is_dir,
    _is_file,
    _is_instantiatable,
    _is_list,
    _join_path,
    _Keys,
    _load_dotenv,
    _load,
    _merge,
    _methods,
    _mkdir,
    _nvidia_smi,
    _osenv,
    _partial,
    _path,
    _pipe,
    _print,
    _records_to_dataframe,
    _run,
    _save,
    _select,
    _set_cuda,
    _setLogger,
    _SPLITS,
    _stop_env_,
    _to_config,
    _to_container,
    _to_dateparm,
    _to_datetime,
    _to_dict,
    _to_numeric,
    _to_yaml,
    _update,
    _viewsource,
    DictKeyType,
)
from ekorpkit.io.google import _mount_google_drive
from ekorpkit.config import Environments


logger = _getLogger(__name__)


class eKonf:
    """ekorpkit config primary class"""

    __version__ = __version__()
    __ekorpkit_path__ = __ekorpkit_path__()
    __home_path__ = __home_path__()
    config = _config
    book_repo = "https://github.com/entelecheia/ekorpkit-book/raw/main/"
    book_repo_assets = book_repo + "assets/"
    book_url = "https://entelecheia.github.io/ekorpkit-book/"
    book_assets_url = book_url + "assets/"
    Keys = _Keys
    Defaults = _Defaults
    SPLITS = _SPLITS

    def __init__(self) -> None:
        raise NotImplementedError("Use one of the static construction functions")

    @staticmethod
    def env() -> Environments:
        """Return the current environments"""
        return Environments()

    @staticmethod
    def compose(
        config_group: str = None,
        overrides: List[str] = [],
        *,
        return_as_dict: bool = False,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        config_name="ekonf",
        verbose: bool = False,
    ):
        return _compose(
            config_group=config_group,
            overrides=overrides,
            return_as_dict=return_as_dict,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
            config_name=config_name,
            verbose=verbose,
        )

    @staticmethod
    def select(
        cfg: Any,
        key: str,
        *,
        default: Any = None,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
    ):
        return _select(
            cfg,
            key,
            default=default,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
        )

    @staticmethod
    def to_dict(
        cfg: Any,
    ):
        return _to_dict(cfg)

    @staticmethod
    def to_config(
        cfg: Any,
    ):
        return _to_config(cfg)

    @staticmethod
    def to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
        if resolve:
            cfg = _to_dict(cfg)
        return _to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)

    @staticmethod
    def to_container(
        cfg: Any,
        *,
        resolve: bool = False,
        throw_on_missing: bool = False,
        enum_to_str: bool = False,
        structured_config_mode: SCMode = SCMode.DICT,
    ):
        return _to_container(
            cfg,
            resolve,
            throw_on_missing,
            enum_to_str,
            structured_config_mode,
        )

    @staticmethod
    def partial(
        config: Any = None, config_group: str = None, *args: Any, **kwargs: Any
    ) -> Any:
        return _partial(config=config, config_group=config_group, *args, **kwargs)

    @staticmethod
    def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
        return _instantiate(config, *args, **kwargs)

    @staticmethod
    def is_config(
        cfg: Any,
    ):
        return _is_config(cfg)

    @staticmethod
    def is_list(
        cfg: Any,
    ):
        return _is_list(cfg)

    @staticmethod
    def is_instantiatable(cfg: Any):
        return _is_instantiatable(cfg)

    @staticmethod
    def load(file_: Union[str, Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
        return _load(file_)

    @staticmethod
    def update(_dict, _overrides):
        return _update(_dict, _overrides)

    @staticmethod
    def merge(
        *configs: Union[
            DictConfig,
            ListConfig,
            Dict[DictKeyType, Any],
            List[Any],
            Tuple[Any, ...],
            Any,
        ],
    ) -> Union[ListConfig, DictConfig]:
        """
        Merge a list of previously created configs into a single one
        :param configs: Input configs
        :return: the merged config object.
        """
        return _merge(*configs)

    @staticmethod
    def save(config: Any, f: Union[str, Path, IO[Any]], resolve: bool = False) -> None:
        _save(config, f, resolve)

    @staticmethod
    def pprint(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def print(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def methods(cfg: Any, obj: object, return_function=False):
        _methods(cfg, obj, return_function)

    @staticmethod
    def function(cfg: Any, _name_, return_function=False, **parms):
        return _function(cfg, _name_, return_function, **parms)

    @staticmethod
    def run(config: Any, **kwargs: Any) -> Any:
        _run(config, **kwargs)

    @staticmethod
    def load_dotenv(verbose: bool = False):
        _load_dotenv(verbose)

    @staticmethod
    def _init_env_(cfg, verbose=False):
        _init_env_(cfg, verbose=verbose)

    @staticmethod
    def _stop_env_(cfg, verbose=False):
        _stop_env_(cfg, verbose=verbose)

    @staticmethod
    def path(
        url_or_filename,
        extract_archive: bool = False,
        force_extract: bool = False,
        return_parent_dir: bool = False,
        cache_dir=None,
        verbose: bool = False,
    ):
        """
        Given something that might be a URL or local path, determine which.
        If it's a remote resource, download the file and cache it, and
        then return the path to the cached file. If it's already a local path,
        make sure the file exists and return the path.

        For URLs, the following schemes are all supported out-of-the-box:

        * ``http`` and ``https``,
        * ``s3`` for objects on `AWS S3`_,
        * ``gs`` for objects on `Google Cloud Storage (GCS)`_, and
        * ``gd`` for objects on `Google Drive`_, and
        * ``hf`` for objects or repositories on `HuggingFace Hub`_.

        Examples
        --------

        To download a file over ``https``::

            cached_path("https://github.com/allenai/cached_path/blob/main/README.md")

        To download an object on GCS::

            cached_path("gs://allennlp-public-models/lerc-2020-11-18.tar.gz")

        To download the PyTorch weights for the model `epwalsh/bert-xsmall-dummy`_
        on HuggingFace, you could do::

            cached_path("hf://epwalsh/bert-xsmall-dummy/pytorch_model.bin")

        For paths or URLs that point to a tarfile or zipfile, you can append the path
        to a specific file within the archive to the ``url_or_filename``, preceeded by a "!".
        The archive will be automatically extracted (provided you set ``extract_archive`` to ``True``),
        returning the local path to the specific file. For example::

            cached_path("model.tar.gz!weights.th", extract_archive=True)

        .. _epwalsh/bert-xsmall-dummy: https://huggingface.co/epwalsh/bert-xsmall-dummy

        Parameters
        ----------

        url_or_filename :
            A URL or path to parse and possibly download.

        extract_archive :
            If ``True``, then zip or tar.gz archives will be automatically extracted.
            In which case the directory is returned.

        force_extract :
            If ``True`` and the file is an archive file, it will be extracted regardless
            of whether or not the extracted directory already exists.

            .. caution::
                Use this flag with caution! This can lead to race conditions if used
                from multiple processes on the same file.

        cache_dir :
            The directory to cache downloads. If not specified, the global default cache directory
            will be used (``~/.cache/cached_path``). This can be set to something else with
            :func:`set_cache_dir()`.

        Returns
        -------
        :class:`pathlib.Path`
            The local path to the (potentially cached) resource.

        Raises
        ------
        ``FileNotFoundError``

            If the resource cannot be found locally or remotely.

        ``ValueError``
            When the URL is invalid.

        ``Other errors``
            Other error types are possible as well depending on the client used to fetch
            the resource.

        """
        return _path(
            url_or_filename,
            extract_archive=extract_archive,
            force_extract=force_extract,
            return_parent_dir=return_parent_dir,
            cache_dir=cache_dir,
            verbose=verbose,
        )

    @staticmethod
    def pipe(data=None, cfg=None):
        return _pipe(data, cfg)

    @staticmethod
    def dependencies(_key=None, path=None):
        return _dependencies(_key, path)

    @staticmethod
    def ensure_list(value):
        return _ensure_list(value)

    @staticmethod
    def to_dateparm(_date, _format="%Y-%m-%d"):
        return _to_dateparm(_date, _format)

    @staticmethod
    def exists(a, *p):
        return _exists(a, *p)

    @staticmethod
    def is_file(a, *p):
        return _is_file(a, *p)

    @staticmethod
    def is_dir(a, *p):
        return _is_dir(a, *p)

    @staticmethod
    def mkdir(_path: str):
        return _mkdir(_path)

    @staticmethod
    def join_path(a, *p):
        return _join_path(a, *p)

    @staticmethod
    def apply(
        func,
        series,
        description=None,
        use_batcher=True,
        minibatch_size=None,
        num_workers=None,
        verbose=False,
        **kwargs,
    ):
        return _apply(
            func,
            series,
            description=description,
            use_batcher=use_batcher,
            minibatch_size=minibatch_size,
            num_workers=num_workers,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def ensure_kwargs(_kwargs, _fn):
        return _ensure_kwargs(_kwargs, _fn)

    @staticmethod
    def save_data(
        data,
        filename=None,
        base_dir=None,
        columns=None,
        index=False,
        filetype="parquet",
        suffix=None,
        verbose=False,
        **kwargs,
    ):
        from ekorpkit.io.file import save_data

        if filename is None:
            raise ValueError("filename must be specified")
        save_data(
            data,
            filename,
            base_dir=base_dir,
            columns=columns,
            index=index,
            filetype=filetype,
            suffix=suffix,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def load_data(filename=None, base_dir=None, verbose=False, **kwargs):
        from ekorpkit.io.file import load_data

        if _Keys.TARGET in kwargs:
            return eKonf.instantiate(
                kwargs, filename=filename, base_dir=base_dir, verbose=verbose
            )
        else:
            if filename is None:
                raise ValueError("filename must be specified")
            return load_data(filename, base_dir=base_dir, verbose=verbose, **kwargs)

    @staticmethod
    def get_filepaths(
        filename_patterns=None, base_dir=None, recursive=True, verbose=True, **kwargs
    ):
        from ekorpkit.io.file import get_filepaths

        if filename_patterns is None:
            raise ValueError("filename must be specified")
        return get_filepaths(
            filename_patterns,
            base_dir=base_dir,
            recursive=recursive,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def concat_data(
        data,
        columns=None,
        add_key_as_name=False,
        name_column="_name_",
        ignore_index=True,
        verbose=False,
        **kwargs,
    ):
        from ekorpkit.io.file import concat_data

        return concat_data(
            data,
            columns=columns,
            add_key_as_name=add_key_as_name,
            name_column=name_column,
            ignore_index=ignore_index,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def is_dataframe(data):
        from ekorpkit.io.file import is_dataframe

        return is_dataframe(data)

    @staticmethod
    def is_colab():
        return _is_colab()

    @staticmethod
    def is_notebook():
        return _is_notebook()

    @staticmethod
    def osenv(key):
        return _osenv(key)

    @staticmethod
    def env_set(key, value):
        return _env_set(key, value)

    @staticmethod
    def nvidia_smi():
        return _nvidia_smi()

    @staticmethod
    def ensure_import_module(name, libpath, liburi, specname=None, syspath=None):
        from ekorpkit.utils.lib import ensure_import_module

        return ensure_import_module(name, libpath, liburi, specname, syspath)

    @staticmethod
    def collage(
        image_filepaths=None,
        filename_patterns=None,
        base_dir=None,
        output_filepath=None,
        ncols=2,
        num_images=None,
        figsize=(30, 20),
        dpi=300,
        title=None,
        title_fontsize=12,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=12,
        fontcolor=None,
        xlabel=None,
        ylabel=None,
        xticklabels=None,
        yticklabels=None,
        xlabel_fontsize=12,
        ylabel_fontsize=12,
        **kwargs,
    ):
        from ekorpkit.visualize.collage import collage as _collage

        return _collage(
            image_filepaths=image_filepaths,
            filename_patterns=filename_patterns,
            base_dir=base_dir,
            output_filepath=output_filepath,
            ncols=ncols,
            num_images=num_images,
            figsize=figsize,
            dpi=dpi,
            title=title,
            title_fontsize=title_fontsize,
            show_filename=show_filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
            xlabel=xlabel,
            ylabel=ylabel,
            xticklabels=xticklabels,
            yticklabels=yticklabels,
            xlabel_fontsize=xlabel_fontsize,
            ylabel_fontsize=ylabel_fontsize,
            **kwargs,
        )

    @staticmethod
    def make_gif(
        image_filepaths=None,
        filename_patterns=None,
        base_dir=None,
        output_filepath=None,
        duration=100,
        loop=0,
        width=None,
        optimize=True,
        quality=50,
        show=False,
        force=False,
        **kwargs,
    ):
        from ekorpkit.visualize.motion import make_gif as _make_gif

        return _make_gif(
            image_filepaths=image_filepaths,
            filename_patterns=filename_patterns,
            base_dir=base_dir,
            output_filepath=output_filepath,
            duration=duration,
            loop=loop,
            width=width,
            optimize=optimize,
            quality=quality,
            show=show,
            force=force,
            **kwargs,
        )

    @staticmethod
    def to_datetime(data, _format=None, _columns=None, **kwargs):
        return _to_datetime(data, _format, _columns, **kwargs)

    @staticmethod
    def to_numeric(data, _columns=None, errors="coerce", downcast=None, **kwargs):
        return _to_numeric(data, _columns, errors, downcast, **kwargs)

    @staticmethod
    def getLogger(
        name=None,
        log_level=None,
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    ):
        return _getLogger(name, log_level, fmt)

    @staticmethod
    def setLogger(level=None, force=True, **kwargs):
        return _setLogger(level, force, **kwargs)

    @staticmethod
    def set_cuda(device=0):
        return _set_cuda(device)

    @staticmethod
    def mount_google_drive(
        workspace=None,
        project=None,
        mountpoint="/content/drive",
        force_remount=False,
        timeout_ms=120000,
    ):
        return _mount_google_drive(
            workspace, project, mountpoint, force_remount, timeout_ms
        )

    @staticmethod
    def getsource(obj):
        return _getsource(obj)

    @staticmethod
    def viewsource(obj):
        return _viewsource(obj)

    @staticmethod
    def clear_output(wait=False):
        return _clear_output(wait)

    @staticmethod
    def display(
        *objs,
        include=None,
        exclude=None,
        metadata=None,
        transient=None,
        display_id=None,
        raw=False,
        clear=False,
        **kwargs,
    ):
        return _display(
            *objs,
            include=include,
            exclude=exclude,
            metadata=metadata,
            transient=transient,
            display_id=display_id,
            raw=raw,
            clear=clear,
            **kwargs,
        )

    @staticmethod
    def display_image(
        data=None,
        url=None,
        filename=None,
        format=None,
        embed=None,
        width=None,
        height=None,
        retina=False,
        unconfined=False,
        metadata=None,
        **kwargs,
    ):
        return _display_image(
            data=data,
            url=url,
            filename=filename,
            format=format,
            embed=embed,
            width=width,
            height=height,
            retina=retina,
            unconfined=unconfined,
            metadata=metadata,
            **kwargs,
        )

    @staticmethod
    def pip(
        name,
        upgrade=False,
        prelease=False,
        editable=False,
        quiet=False,
        find_links=None,
        requirement=None,
        force_reinstall=False,
        verbose=False,
        **kwargs,
    ):
        from ekorpkit.utils.lib import pip as _pip

        return _pip(
            name,
            upgrade,
            prelease,
            editable,
            quiet,
            find_links,
            requirement,
            force_reinstall,
            verbose,
            **kwargs,
        )

    @staticmethod
    def upgrade(prelease=False, quiet=False, force_reinstall=False, **kwargs):
        from ekorpkit.utils.lib import pip

        return pip(
            name="ekorpkit",
            upgrade=True,
            prelease=prelease,
            quiet=quiet,
            force_reinstall=force_reinstall,
            **kwargs,
        )

    @staticmethod
    def dict_product(dicts):
        return _dict_product(dicts)

    @staticmethod
    def get_display():
        return _get_display()

    @staticmethod
    def hide_code_in_slideshow():
        return _hide_code_in_slideshow()

    @staticmethod
    def cprint(str_color_tuples, **kwargs):
        return _cprint(str_color_tuples)

    @staticmethod
    def dict_to_dataframe(data, orient="columns", dtype=None, columns=None):
        return _dict_to_dataframe(data, orient, dtype, columns)

    @staticmethod
    def records_to_dataframe(
        data, index=None, exclude=None, columns=None, coerce_float=False, nrows=None
    ):
        return _records_to_dataframe(data, index, exclude, columns, coerce_float, nrows)

    @staticmethod
    def create_dropdown(
        options,
        value,
        description,
        disabled=False,
        style={"description_width": "initial"},
        layout=None,
        **kwargs,
    ):
        return _create_dropdown(
            options,
            value,
            description,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_textarea(
        value,
        description,
        placeholder="",
        disabled=False,
        style={"description_width": "initial"},
        layout=None,
        **kwargs,
    ):
        return _create_textarea(
            value,
            description,
            placeholder,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_button(
        description, button_style="", icon="check", layout=None, **kwargs
    ):
        return _create_button(description, button_style, icon, layout, **kwargs)

    @staticmethod
    def create_radiobutton(
        options,
        description,
        value=None,
        disabled=False,
        style={"description_width": "initial"},
        layout=None,
        **kwargs,
    ):
        return _create_radiobutton(
            options,
            description,
            value,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_image(
        filename=None,
        format=None,
        width=None,
        height=None,
        **kwargs,
    ):
        return _create_image(filename, format, width, height, **kwargs)

    @staticmethod
    def create_floatslider(
        min=0.0,
        max=1.0,
        step=0.1,
        value=None,
        description="",
        disabled=False,
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format=".1f",
        style={"description_width": "initial"},
        layout=None,
        **kwargs,
    ):
        return _create_floatslider(
            min,
            max,
            step,
            value,
            description,
            disabled,
            continuous_update,
            orientation,
            readout,
            readout_format,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def get_imagefont(fontname=None, fontsize=12):
        from ekorpkit.visualize.collage import _get_imagefont

        return _get_imagefont(fontname, fontsize)

    @staticmethod
    def read(uri, mode="rb", encoding=None, **kwargs):
        return _read(uri, mode, encoding, **kwargs)



import tensorflow as tf
import tensorflow_io as tfio


def get_dataset(path_to_files):
    autotune = tf.data.experimental.AUTOTUNE
    filenames = tf.data.Dataset.list_files(path_to_files + "/*", shuffle=True)

    def parquet_ds(file):
        ds = tfio.IODataset.from_parquet(file, {"image": tf.string, "label": tf.int32})
        return ds

    ds = filenames.interleave(
        parquet_ds, num_parallel_calls=autotune, deterministic=False
    )

    def parse(example):
        image = tf.io.decode_raw(example["image"], tf.uint8)
        image = tf.reshape(image, [32, 32, 3])
        label = example["label"]
        return image, label

    ds = ds.map(parse, num_parallel_calls=autotune)

    return ds



import os
import logging
import gdown
import shutil
from pathlib import Path
from ekorpkit.io.fetch.cached_path import _cached_path as _cpath

# import cached_path as _cpath

log = logging.getLogger(__name__)


def cached_path(
    url_or_filename,
    extract_archive: bool = False,
    force_extract: bool = False,
    return_parent_dir: bool = False,
    cache_dir=None,
    verbose: bool = False,
):
    if url_or_filename is None:
        return None
    if verbose:
        log.info(
            "caching path: {}, extract_archive: {}, force_extract: {}, cache_dir: {}".format(
                url_or_filename, extract_archive, force_extract, cache_dir
            )
        )

    try:
        if url_or_filename.startswith("gd://"):

            _path = cached_gdown(
                url_or_filename,
                verbose=verbose,
                extract_archive=extract_archive,
                force_extract=force_extract,
                cache_dir=cache_dir,
            )
        else:
            if cache_dir is None:
                cache_dir = Path.home() / ".ekorpkit" / ".cache" / "cached_path"
            else:
                cache_dir = Path(cache_dir) / "cached_path"

            _path = _cpath.cached_path(
                url_or_filename,
                extract_archive=extract_archive,
                force_extract=force_extract,
                cache_dir=cache_dir,
            ).as_posix()

        log.info(f"cached path: {_path}")

        if Path(_path).is_file():
            _parent_dir = Path(_path).parent
        elif Path(_path).is_dir():
            _parent_dir = Path(_path)
        else:
            log.warning(f"Unknown path: {_path}")
            return None

        if return_parent_dir:
            return _parent_dir.as_posix()
        else:
            return _path

    except Exception as e:
        log.error(e)
        return None


def cached_gdown(
    url, verbose=False, extract_archive=None, force_extract=False, cache_dir=None
):
    """
    :type url: str
          ex) gd://id:path
    :type verbose: bool
    :type extract_archive: bool
    :type force_extract: bool
    :type cache_dir: str
    :returns: str
    """

    if verbose:
        log.info(f"Downloading {url}...")
    if cache_dir is None:
        cache_dir = Path.home() / ".ekorpkit" / ".cache" / "gdown"
    else:
        cache_dir = Path(cache_dir) / "gdown"
    cache_dir.mkdir(parents=True, exist_ok=True)

    gd_prefix = "gd://"
    if url.startswith(gd_prefix):
        url = url[len(gd_prefix) :]
        _url = url.split(":")
        if len(_url) == 2:
            id, path = _url
        else:
            id = _url[0]
            path = id

        # If we're using the path!c/d/file.txt syntax, handle it here.
        fname = None
        extraction_path = path
        exclamation_index = path.find("!")
        if extract_archive and exclamation_index >= 0:
            extraction_path = path[:exclamation_index]
            fname = path[exclamation_index + 1 :]

        cache_path = cache_dir / f".{id}" / extraction_path
        cache_path.parent.mkdir(parents=True, exist_ok=True)

        cache_path = gdown.cached_download(
            id=id,
            path=cache_path.as_posix(),
            quiet=not verbose,
        )

        if extract_archive:
            extraction_path, files = extractall(cache_path, force_extract=force_extract)

            if fname and files:
                for f in files:
                    if f.endswith(fname):
                        return f
            else:
                return extraction_path
        return cache_path

    else:
        log.warning(f"Unknown url: {url}")
        return None


def extractall(path, to=None, force_extract=False):
    """Extract archive file.

    Parameters
    ----------
    path: str
        Path of archive file to be extracted.
    to: str, optional
        Directory to which the archive file will be extracted.
        If None, it will be set to the parent directory of the archive file.
    """
    import tarfile
    from zipfile import ZipFile

    if to is None:
        to = os.path.dirname(path)

    if path.endswith(".zip"):
        opener, mode = ZipFile, "r"
    elif path.endswith(".tar"):
        opener, mode = tarfile.open, "r"
    elif path.endswith(".tar.gz") or path.endswith(".tgz"):
        opener, mode = tarfile.open, "r:gz"
    elif path.endswith(".tar.bz2") or path.endswith(".tbz"):
        opener, mode = tarfile.open, "r:bz2"
    else:
        log.warning(
            "Could not extract '%s' as no appropriate " "extractor is found" % path
        )
        return path, None

    def namelist(f):
        if isinstance(f, ZipFile):
            return f.namelist()
        return [m.path for m in f.members]

    def filelist(f):
        files = []
        for fname in namelist(f):
            fname = os.path.join(to, fname)
            files.append(fname)
        return files

    extraction_name = Path(path).stem
    extraction_path = f"{to}/{extraction_name}"
    if extraction_path is not None:
        # If the extracted directory already exists (and is non-empty), then no
        # need to extract again unless `force_extract=True`.
        if (
            os.path.isdir(extraction_path)
            and os.listdir(extraction_path)
            and not force_extract
        ):
            files = [
                os.path.join(dirpath, filename)
                for dirpath, _, filenames in os.walk(extraction_path)
                for filename in filenames
            ]

            return to, files

    with opener(path, mode) as f:
        f.extractall(path=to)

    return to, filelist(f)



import logging
import pandas as pd
import numpy as np
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from .base import BaseSentimentAnalyser, _Keys


log = logging.getLogger(__name__)


class SentimentAnalyser(BaseSentimentAnalyser):
    def __init__(self, **args):
        args = eKonf.to_config(args)
        super().__init__(**args)

        self._ngram = eKonf.instantiate(args.model.ngram)

    def analyze(self, text, **kwargs):
        """
        :type text: str
        :returns: list
        """
        return self._ngram.find_features(text, **kwargs)

    def tokenize(self, text):
        """Tokenize text.

        :returns: list
        """
        return self._ngram.tokenize(text)

    def predict(self, data, _predict_={}):
        if _predict_:
            self._predict_ = _predict_

        input_col = self._predict_[eKonf.Keys.INPUT]
        _method_ = self._predict_[eKonf.Keys.METHOD]
        _meth_name_ = _method_.get(eKonf.Keys.METHOD_NAME)
        log.info(
            f"Predicting sentiments of the column [{input_col}] using {_meth_name_}"
        )
        _fn = lambda doc: getattr(self, _meth_name_)(doc)
        with elapsed_timer(format_time=True) as elapsed:
            predictions = eKonf.apply(
                _fn,
                data[input_col],
                description=f"Predicting [{input_col}]",
                verbose=self.verbose,
                use_batcher=self.use_batcher,
                num_workers=self.num_workers,
            )
            pred_df = pd.DataFrame(predictions.tolist(), index=predictions.index)
            data = data.join(pred_df)
            log.info(" >> elapsed time to predict: {}".format(elapsed()))

        return data

    def predict_sentence(self, text, features="polarity", min_examples=5):
        """Get score for a list of terms.

        :returns: dict
        """
        scores = {}

        features = self._predict_.features or features
        features = eKonf.ensure_list(features)
        _lex_feats = self._predict_.get(_Keys.LEXICON_FEATURES)
        min_examples = self._predict_.min_tokens or min_examples

        tokens = self._ngram.tokenize(text)
        num_examples = len(tokens)
        lexicon_features = self._ngram.find_features(tokens, features=_lex_feats)
        for feature in features:
            if num_examples < min_examples:
                score = {feature: np.nan}
            else:
                score = self._get_score(
                    lexicon_features,
                    feature=feature,
                    num_examples=num_examples,
                    min_examples=min_examples,
                )
            score = self._assign_class(score, feature=feature)
            scores.update(score)

        return scores

    def _get_score(
        self, lexicon_features, feature="polarity", num_examples=None, min_examples=5
    ):
        """Get score for features.

        :returns: dict
        """
        _features = self._features
        _feature_method = _features[feature]
        _default_method = self._features[_Keys.DEFAULT]
        _lex_feats = _feature_method.get(_Keys.LEXICON_FEATURES)
        _num_examples = _feature_method.get(_Keys.NUM_EXAMPLES) or _Keys.NUM_TOKENS.value

        lxfeat = pd.DataFrame.from_dict(lexicon_features, orient="index")

        score = {_num_examples: num_examples, feature: np.nan}
        if lxfeat.empty:
            num_examples = 0
        elif num_examples is None:
            num_examples = lxfeat.shape[0]
        score[_num_examples] = num_examples

        if num_examples > min_examples:

            eps = self.EPSILON
            if _Keys.CONDITIONS in _feature_method:
                _conditions = _feature_method.get(_Keys.CONDITIONS)
                _count = _feature_method.get(_Keys.COUNT)
                _aggs = _feature_method.get(_Keys.AGGS)
                _score = _feature_method.get(_Keys.SCORE)
            else:
                _conditions = _default_method.get(_Keys.CONDITIONS)
                _count = _default_method.get(_Keys.COUNT)
                _aggs = _default_method.get(_Keys.AGGS)
                _score = _default_method.get(_Keys.SCORE)
                lxfeat.rename(columns={_lex_feats: _Keys.FEATURE.value}, inplace=True)
            if self.verbose > 5:
                log.info("Evaluating %s", feature)
                print(lxfeat)
            for _name, _expr in _conditions.items():
                lxfeat[_name] = np.where(
                    lxfeat.eval(_expr), lxfeat[_count] if _count else 1, 0
                )
            _aggs = {k: v for k, v in _aggs.items() if k in lxfeat.columns}
            lxfeat_agg = lxfeat.agg(_aggs)
            lxfeat_agg = pd.DataFrame(lxfeat_agg).T
            _feat_score = lxfeat_agg.eval(_score)
            if not _feat_score.empty:
                score[feature] = _feat_score[0]
            else:
                score[feature] = None

        return score

    def _assign_class(self, score, feature="polarity"):
        """Assign class to a score.

        :returns: dict
        """
        _label_key = feature + "_label"
        _labels = self._features[feature].get(_Keys.LABELS)
        if _labels:
            score[_label_key] = None
            if feature not in score:
                if self.verbose > 5:
                    log.info(f"No score for {feature}")
                return score
            for _label, _thresh in _labels.items():
                if isinstance(_thresh, str):
                    _thresh = eval(_thresh)
                if score[feature] >= _thresh[0] and score[feature] <= _thresh[1]:
                    score[_label_key] = _label
        return score

    def predict_article(
        self,
        article,
        features=["polarity"],
        min_examples=2,
        **kwargs,
    ):
        scores = {}

        features = self._predict_.features or features
        features = eKonf.ensure_list(features)
        min_examples = self._predict_.min_sentences or min_examples

        article_scores = {}
        if isinstance(article, str):
            sents = article.split(self._sentence_separator)
            num_examples = len(sents)
            for sent_no, sent in enumerate(sents):
                sent = sent.strip()
                if sent:
                    _scores = self.predict_sentence(sent, features=features)
                    if _scores:
                        article_scores[sent_no] = _scores
        for feature in features:
            score = self._get_aggregate_scores(
                article_scores,
                feature=feature,
                min_examples=min_examples,
            )
            scores.update(score)

        return scores

    def _get_aggregate_scores(
        self,
        scores,
        feature="polarity",
        num_examples=None,
        min_examples=2,
    ):
        """Get aggreagate score for features.

        :returns: dict
        """
        _article_features = self._article_features
        _agg_method = _article_features.get(feature) or _article_features[_Keys.DEFAULT]
        _num_examples = _agg_method.get(_Keys.NUM_EXAMPLES) or _Keys.NUM_TOKENS.value

        scores_df = pd.DataFrame.from_dict(scores, orient="index")
        scores_df.dropna(subset=[feature], inplace=True)

        eps = self.EPSILON
        _conditions = _agg_method.get(_Keys.CONDITIONS)
        _count = _agg_method.get(_Keys.COUNT)
        _aggs = eKonf.to_dict(_agg_method.get(_Keys.AGGS))
        _evals = _agg_method.get(_Keys.EVALS)
        _scores = _agg_method.get(_Keys.SCORES)
        # _label_by = _agg_method.get(_Keys.LABEL_BY)
        _scores = {
            _name: _name.replace(_Keys.FEATURE.value, feature) for _name in _scores
        }

        score = {_num_examples: np.nan}
        for f in _scores.values():
            score[f] = np.nan

        num_examples = scores_df.shape[0] if not scores_df.empty else 0
        score[_num_examples] = num_examples
        if num_examples > min_examples:

            scores_df.rename(columns={feature: _Keys.FEATURE.value}, inplace=True)
            if self.verbose > 5:
                log.info("Evaluating %s", feature)
                print(scores_df)
            if _conditions is not None:
                for _name, _expr in _conditions.items():
                    scores_df[_name] = np.where(
                        scores_df.eval(_expr), scores_df[_count] if _count else 1, 0
                    )
            _aggs = {k: v for k, v in _aggs.items() if k in scores_df.columns}
            lxfeat_agg = scores_df.agg(_aggs)
            lxfeat_agg = pd.DataFrame(lxfeat_agg).T
            lxfeat_agg = lxfeat_agg.unstack().to_frame().T
            lxfeat_agg.columns = [
                f"{c[1]}_{c[0]}" for c in lxfeat_agg.columns.to_flat_index()
            ]
            if _evals is not None:
                for _name, _expr in _evals.items():
                    lxfeat_agg[_name] = lxfeat_agg.eval(_expr)
            lxfeat_agg.rename(columns=_scores, inplace=True)
            _feat_score = lxfeat_agg[_scores.values()].iloc[0].to_dict()
            if _feat_score:
                score.update(_feat_score)

        for f in _scores.values():
            if f.startswith(feature):
                score = self._assign_article_class(
                    score, feature=feature, article_feature=f
                )

        return score

    def _assign_article_class(self, score, feature="polarity", article_feature=None):
        """Assign class to a score.

        :returns: dict
        """
        _label_key = article_feature + "_label"
        _labels = self._article_features.get(feature, {}).get(_Keys.LABELS)
        if _labels:
            score[_label_key] = None
            if article_feature not in score:
                if self.verbose > 5:
                    log.info(f"No score for {article_feature}")
                return score
            for _label, _thresh in _labels.items():
                if isinstance(_thresh, str):
                    _thresh = eval(_thresh)
                if (
                    score[article_feature] >= _thresh[0]
                    and score[article_feature] <= _thresh[1]
                ):
                    score[_label_key] = _label
        return score



import logging
import pickle
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from ...analysis import Scorer
from ...utils import probs_to_preds


class BaseLabeler(ABC):
    """Abstract baseline label voter class."""

    def __init__(self, cardinality: int = 2, **kwargs: Any) -> None:
        self.cardinality = cardinality

    @abstractmethod
    def predict_proba(self, L: np.ndarray) -> np.ndarray:
        """Abstract method for predicting probabilistic labels given a label matrix.

        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}f

        Returns
        -------
        np.ndarray
            An [n,k] array of probabilistic labels
        """
        pass

    def predict(
        self,
        L: np.ndarray,
        return_probs: Optional[bool] = False,
        tie_break_policy: str = "abstain",
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return predicted labels, with ties broken according to policy.

        Policies to break ties include:
        "abstain": return an abstain vote (-1)
        "true-random": randomly choose among the tied options
        "random": randomly choose among tied option using deterministic hash

        NOTE: if tie_break_policy="true-random", repeated runs may have slightly different
        results due to difference in broken ties


        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}
        return_probs
            Whether to return probs along with preds
        tie_break_policy
            Policy to break ties when converting probabilistic labels to predictions

        Returns
        -------
        np.ndarray
            An [n,1] array of integer labels

        (np.ndarray, np.ndarray)
            An [n,1] array of integer labels and an [n,k] array of probabilistic labels
        """
        Y_probs = self.predict_proba(L)
        Y_p = probs_to_preds(Y_probs, tie_break_policy)
        if return_probs:
            return Y_p, Y_probs
        return Y_p

    def score(
        self,
        L: np.ndarray,
        Y: np.ndarray,
        metrics: Optional[List[str]] = ["accuracy"],
        tie_break_policy: str = "abstain",
    ) -> Dict[str, float]:
        """Calculate one or more scores from user-specified and/or user-defined metrics.

        Parameters
        ----------
        L
            An [n,m] matrix with values in {-1,0,1,...,k-1}
        Y
            Gold labels associated with data points in L
        metrics
            A list of metric names
        tie_break_policy
            Policy to break ties when converting probabilistic labels to predictions


        Returns
        -------
        Dict[str, float]
            A dictionary mapping metric names to metric scores
        """
        if tie_break_policy == "abstain":  # pragma: no cover
            logging.warning(
                "Metrics calculated over data points with non-abstain labels only"
            )

        Y_pred, Y_prob = self.predict(
            L, return_probs=True, tie_break_policy=tie_break_policy
        )

        scorer = Scorer(metrics=metrics)
        results = scorer.score(Y, Y_pred, Y_prob)
        return results

    def save(self, destination: str) -> None:
        """Save label model.

        Parameters
        ----------
        destination
            Filename for saving model

        Example
        -------
        >>> label_model.save('./saved_label_model.pkl')  # doctest: +SKIP
        """
        f = open(destination, "wb")
        pickle.dump(self.__dict__, f)
        f.close()

    def load(self, source: str) -> None:
        """Load existing label model.

        Parameters
        ----------
        source
            Filename to load model from

        Example
        -------
        Load parameters saved in ``saved_label_model``

        >>> label_model.load('./saved_label_model.pkl')  # doctest: +SKIP
        """
        f = open(source, "rb")
        tmp_dict = pickle.load(f)
        f.close()
        self.__dict__.update(tmp_dict)
